// DSWifi Project - sgIP Internet Protocol Stack Implementation
// Copyright (C) 2005 Stephen Stair - sgstair@akkit.org - http://www.akkit.org
/****************************************************************************** 
DSWifi Lib and test materials are licenced under the MIT open source licence:
Copyright (c) 2005 Stephen Stair

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
******************************************************************************/

#include "sgIP_TCP.h"
#include "sgIP_IP.h"
#include "sgIP_Hub.h"

sgIP_Record_TCP * tcprecords;
int port_counter;

void sgIP_TCP_Init() {
	tcprecords=0;
	port_counter=SGIP_TCP_FIRSTOUTGOINGPORT;
}


unsigned long sgIP_TCP_support_seqhash(sgIP_Record_TCP * rec) {
	unsigned long hash;
	hash=rec->destip;
	hash ^= rec->destport * 0x02041089;
	hash ^= rec->srcport * 0x080810422;
	hash ^= rec->srcip * 0x48841221;
	hash ^= rec->destip * 0x04020108;
	return hash;
}
int sgIP_TCP_GetUnusedOutgoingPort() {
	int myport,clear;
	sgIP_Record_TCP * rec = tcprecords;
	while(1) {
		sgIP_Record_TCP * rec = tcprecords;
		myport=port_counter++;
		if(port_counter>SGIP_TCP_LASTOUTGOINGPORT) port_counter=SGIP_TCP_FIRSTOUTGOINGPORT;
		clear=1;
		while(rec) {
			if(rec->srcport==myport && rec->tcpstate!=SGIP_TCP_STATE_CLOSED && rec->tcpstate!=SGIP_TCP_STATE_NODATA) { clear=0; break; }
			rec=rec->next;
		}
		if(clear) return myport;
	}
}

int sgIP_TCP_CalcIncomingChecksum(sgIP_memblock * mb, unsigned long srcip, unsigned long destip, int totallength) {
	int checksum,i;
	if(!mb) return 0;
	if(mb->totallength&1) mb->datastart[mb->totallength]=0;
	checksum=0;
	for(i=0;i<((mb->totallength+1)>>1);i++) { // horribly slow I know... will optimize later.
		checksum+=((unsigned short *)mb->datastart)[i];
		checksum+=(checksum>>16);
		checksum&=0xFFFF;
	}
	// add in checksum of "faux header"
	checksum+=(destip&0xFFFF);
	checksum+=(destip>>16);
	checksum+=(srcip&0xFFFF);
	checksum+=(srcip>>16);
	checksum+=htons(totallength);
	checksum+=(6)<<8;
	checksum+=(checksum>>16);
	checksum&=0xFFFF;

	return checksum;
}


int sgIP_TCP_ReceivePacket(sgIP_memblock * mb, unsigned long srcip, unsigned long destip) {
	if(!mb) return 0;
	sgIP_Header_TCP * tcp;
	tcp = (sgIP_Header_TCP *) mb->datastart;
	//                   01234567890123456789012345678901
	//                   TCPxxxxxxxx-xxxxxxxx,xxxx-xxxx
	SGIP_DEBUG_MESSAGE(("TCP%08X-%08X,%04X-%04X",srcip,destip,tcp->srcport,tcp->destport));
	//                   -Lxxxx,Cxxxx,Fxx,hx,Axxxxxxxx
	SGIP_DEBUG_MESSAGE(("-L%04X,C%04X,F%02X,h%X,A%08X",mb->totallength,tcp->checksum,tcp->tcpflags,tcp->dataofs_>>4,tcp->acknum));
	if(sgIP_TCP_CalcIncomingChecksum(mb,srcip,destip,mb->totallength)!=0xFFFF) { // checksum is invalid!
		SGIP_DEBUG_MESSAGE(("TCP receive checksum incorrect"));
		sgIP_memblock_free(mb);
		return 0;
	}
	sgIP_Record_TCP * rec;
	rec=tcprecords;
	// find associated block.
	while(rec) {
		if(rec->srcport==tcp->destport && rec->srcip==destip) break;
		rec=rec->next;
	}
	if(!rec) {
		sgIP_memblock_free(mb);
		return 0;
	}
	// check sequence and ACK numbers, to ensure they're in range.

	// check extraneous situations, like RST

	// now, decide what to do with our nice new shiny memblock...
	switch(rec->tcpstate) {
	case SGIP_TCP_STATE_NODATA: // newly allocated
	case SGIP_TCP_STATE_UNUSED: // allocated & BINDed
	case SGIP_TCP_STATE_CLOSED: // Block is unused.
		break; // can't do anything in these states.
	case SGIP_TCP_STATE_LISTEN: // listening
		break;
	case SGIP_TCP_STATE_SYN_SENT: // connect initiated

		break;
	case SGIP_TCP_STATE_SYN_RECEIVED: // spawned from listen socket;
		break;
	case SGIP_TCP_STATE_ESTABLISHED: // syns have been exchanged
		break;
	case SGIP_TCP_STATE_FIN_WAIT_1: // sent a FIN, haven't got FIN or ACK yet.
		break;
	case SGIP_TCP_STATE_FIN_WAIT_2: // got ACK for our FIN, haven't got FIN yet.
		break;
	case SGIP_TCP_STATE_CLOSE_WAIT: // got FIN, wait for user code to close socket & send FIN
		break;
	case SGIP_TCP_STATE_CLOSING: // got FIN, waiting for ACK of our FIN
		break;
	case SGIP_TCP_STATE_LAST_ACK: // wait for ACK of our last FIN
		break;
	case SGIP_TCP_STATE_TIME_WAIT: // wait to ensure remote tcp knows it's been terminated.
		break;
	}
	sgIP_memblock_free;
	return 0;
}

sgIP_memblock * sgIP_TCP_GenHeader(sgIP_Record_TCP * rec, int flags, int datalength) {
	sgIP_memblock * mb = sgIP_memblock_alloc(datalength+20+sgIP_IP_RequiredHeaderSize());
	int windowlen;
	if(!mb) return 0;
	sgIP_memblock_exposeheader(mb,-sgIP_IP_RequiredHeaderSize()); // hide IP header space for later
	sgIP_Header_TCP * tcp = (sgIP_Header_TCP *) mb->datastart;
	tcp->srcport=rec->srcport;
	tcp->destport=rec->destport;
	tcp->seqnum=rec->sequence;
	tcp->acknum=rec->ack;
	tcp->tcpflags=flags;
	tcp->urg_ptr=0; // no support for URG data atm.
	tcp->checksum=0;
	tcp->dataofs_=5<<4; // header length == 20 (5*32bit)
	windowlen=rec->buf_rx_out-rec->buf_rx_in;
	if(windowlen<0) windowlen+=SGIP_TCP_RECEIVEBUFFERLENGTH; // we now have the amount in the buffer
	windowlen = SGIP_TCP_RECEIVEBUFFERLENGTH-windowlen-1;
	if(windowlen<0) windowlen=0;
	if(windowlen>65535) windowlen=65535;
	tcp->window=windowlen;
	return mb;
}
void sgIP_TCP_CalcChecksum(sgIP_Record_TCP * rec, sgIP_memblock * mb) {
	int checksum,i;
	if(!rec || !mb) return;
	if(mb->totallength&1) mb->datastart[mb->totallength]=0;
	checksum=0;
	for(i=0;i<((mb->totallength+1)>>1);i++) { // horribly slow I know... will optimize later.
		checksum+=((unsigned short *)mb->datastart)[i];
		checksum+=(checksum>>16);
		checksum&=0xFFFF;
	}
	// add in checksum of "faux header"
	checksum+=(rec->destip&0xFFFF);
	checksum+=(rec->destip>>16);
	checksum+=(rec->srcip&0xFFFF);
	checksum+=(rec->srcip>>16);
	checksum+=htons(mb->totallength);
	checksum+=(6)<<8;
	checksum+=(checksum>>16);
	checksum&=0xFFFF;

	checksum = ~checksum;
	if(checksum==0) checksum=0xFFFF;

	sgIP_Header_TCP * tcp;
	tcp = (sgIP_Header_TCP *) mb->datastart;

	tcp->checksum=checksum;
}
int sgIP_TCP_SendPacketDirect(sgIP_Record_TCP * rec, int flags, char * data, int datalength) { // data sent is taken from passed variables.
	if(!rec) return 0; // error error!
	SGIP_INTR_PROTECT();
	sgIP_memblock * mb =sgIP_TCP_GenHeader(rec,flags,datalength);
	if(!mb) {
		SGIP_INTR_UNPROTECT();
		return 0; // oh noes!
	}
	int i;
	for(i=0;i<datalength;i++) {
		mb->datastart[i+20]=data[i];
	}
	sgIP_TCP_CalcChecksum(rec,mb);
	sgIP_IP_SendViaIP(mb,6,rec->srcip,rec->destip);
	SGIP_INTR_UNPROTECT();
}
int sgIP_TCP_SendPacket(sgIP_Record_TCP * rec, int flags, int datalength) { // data sent is taken directly from the TX fifo.
	if(!rec) return 0;
	SGIP_INTR_PROTECT();
	sgIP_memblock * mb =sgIP_TCP_GenHeader(rec,flags,datalength);
	if(!mb) {
		SGIP_INTR_UNPROTECT();
		return 0;
	}
	int i,j;
	j=rec->buf_tx_in;
	for(i=0;i<datalength;i++) {
		mb->datastart[i+20]=rec->buf_tx[j++];
		if(j==SGIP_TCP_TRANSMITBUFFERLENGTH) j=0;
	}
	sgIP_TCP_CalcChecksum(rec,mb);
	sgIP_IP_SendViaIP(mb,6,rec->srcip,rec->destip);
	SGIP_INTR_UNPROTECT();
}

sgIP_Record_TCP * sgIP_TCP_AllocRecord() {
	SGIP_INTR_PROTECT();
	sgIP_Record_TCP * rec;
	rec = sgIP_malloc(sizeof(sgIP_Record_TCP));
	if(rec) {
		rec->buf_oob_in=0;
		rec->buf_oob_out=0;
		rec->buf_rx_in=0;
		rec->buf_rx_out=0;
		rec->buf_tx_in=0;
		rec->buf_tx_out=0;
		rec->tcpstate=0;
		rec->next=tcprecords;
		tcprecords=rec;
		rec->maxlisten=0;
		rec->srcip=0;
	}
	SGIP_INTR_UNPROTECT();
	return rec;
}
void sgIP_TCP_FreeRecord(sgIP_Record_TCP * rec) {
	if(!rec) return;
	SGIP_INTR_PROTECT();
	sgIP_Record_TCP * t;

	rec->tcpstate=0;
	if(tcprecords==rec) {
		tcprecords=rec->next;
	} else {
		t=tcprecords;
		while(t) {
			if(t->next==rec) {
				t->next=rec->next;
				break;
			}
			t=t->next;
		}
	}
	if(rec->listendata) sgIP_free(rec->listendata);
	sgIP_free(rec);

	SGIP_INTR_UNPROTECT();
}

int sgIP_TCP_Bind(sgIP_Record_TCP * rec, int srcport, int destport, unsigned long srcip, unsigned long destip) {
	if(!rec) return 0;
	SGIP_INTR_PROTECT();
	if(rec->tcpstate==SGIP_TCP_STATE_NODATA) {
		rec->srcip=sgIP_IP_GetLocalBindAddr(srcip,destip);
		rec->destip=destip;
		rec->srcport=srcport;
		rec->destport=destport;
		if(rec->tcpstate==SGIP_TCP_STATE_NODATA) rec->tcpstate=SGIP_TCP_STATE_UNUSED;
	}
	SGIP_INTR_UNPROTECT();
	return 0;
}
int sgIP_TCP_Listen(sgIP_Record_TCP * rec, int maxlisten) {
	if(!rec) return 0;
	int err;
	SGIP_INTR_PROTECT();
	if(rec->maxlisten!=0) { // we're *already* listening.
		err=0;
	} else {
		err=0;
		rec->maxlisten=maxlisten;
		rec->listendata = sgIP_malloc(maxlisten*sizeof(sgIP_TCP_ListenStruct));
		if(!rec->listendata) { rec->maxlisten=0; err=0; } else {rec->tcpstate=SGIP_TCP_STATE_LISTEN;}
	}
	SGIP_INTR_UNPROTECT();
	return err;
}
int sgIP_TCP_Close(sgIP_Record_TCP * rec) {
	if(!rec) return 0;
	SGIP_INTR_PROTECT();
	switch(rec->tcpstate) {
	case SGIP_TCP_STATE_NODATA:
	case SGIP_TCP_STATE_UNUSED:
	case SGIP_TCP_STATE_CLOSED: // Block is unused.
	case SGIP_TCP_STATE_TIME_WAIT: // wait to ensure remote tcp knows it's been terminated.
	case SGIP_TCP_STATE_LISTEN: // listening
	case SGIP_TCP_STATE_LAST_ACK: // wait for ACK of our last FIN
	case SGIP_TCP_STATE_CLOSING: // got FIN, waiting for ACK of our FIN
	case SGIP_TCP_STATE_FIN_WAIT_2: // got ACK for our FIN, haven't got FIN yet.
	case SGIP_TCP_STATE_FIN_WAIT_1: // sent a FIN, haven't got FIN or ACK yet.
		break; // do nothing

	case SGIP_TCP_STATE_SYN_SENT: // connect initiated
	case SGIP_TCP_STATE_SYN_RECEIVED: // spawned from listen socket;
		// just close the connection, pretend it never happened. Any additional incoming packets will generate a RST.
		rec->tcpstate=SGIP_TCP_STATE_CLOSED;
		break;

	case SGIP_TCP_STATE_ESTABLISHED: // syns have been exchanged
		// send FIN
		sgIP_TCP_SendPacket(rec,SGIP_TCP_FLAG_FIN | SGIP_TCP_FLAG_ACK, 0);
		rec->tcpstate=SGIP_TCP_STATE_FIN_WAIT_1;
		break;

	case SGIP_TCP_STATE_CLOSE_WAIT: // got FIN, wait for user code to close socket & send FIN
		// send FIN
		sgIP_TCP_SendPacket(rec,SGIP_TCP_FLAG_FIN | SGIP_TCP_FLAG_ACK, 0);
		rec->tcpstate=SGIP_TCP_STATE_LAST_ACK;
		break;
	}

	SGIP_INTR_UNPROTECT();
	return 0;
}
int sgIP_TCP_Connect(sgIP_Record_TCP * rec, unsigned long destip, int destport) {
	if(!rec) return 0;
	SGIP_INTR_PROTECT();
	if(rec->tcpstate==SGIP_TCP_STATE_NODATA) { // need to bind a local address
		rec->srcip=sgIP_IP_GetLocalBindAddr(0,destip);
		rec->srcport=sgIP_TCP_GetUnusedOutgoingPort();
		rec->destip=destip;
		rec->destport=destport;		
	} else if(rec->tcpstate==SGIP_TCP_STATE_UNUSED) { // already bound to a local address.
		rec->destip=destip;
		rec->destport=destport;
	} else {
		SGIP_INTR_UNPROTECT();
		return 0;
	}

	// send a SYN packet, and advance the state of the connection
	rec->sequence=sgIP_TCP_support_seqhash(rec);
	sgIP_TCP_SendPacket(rec,SGIP_TCP_FLAG_SYN,0);
	rec->tcpstate=SGIP_TCP_STATE_SYN_SENT;

	SGIP_INTR_UNPROTECT();
	return 0;
}
int sgIP_TCP_Send(sgIP_Record_TCP * rec, char * datatosend, int datalength, int flags) {
	if(!rec || !datatosend) return 0;
	SGIP_INTR_PROTECT();
	int bufsize;
	bufsize=rec->buf_tx_out-rec->buf_tx_in;
	if(bufsize<0) bufsize+=SGIP_TCP_TRANSMITBUFFERLENGTH;
	bufsize=SGIP_TCP_TRANSMITBUFFERLENGTH-bufsize-1; // space left in buffer
	if(datalength>bufsize) datalength=bufsize;
	int i,j;
	j=rec->buf_tx_out;
	for(i=0;i<datalength;i++) {
		rec->buf_tx[j++]=datatosend[i];
		if(j==SGIP_TCP_TRANSMITBUFFERLENGTH) j=0;
	}
	rec->buf_tx_out = j;
	SGIP_INTR_UNPROTECT();
	return datalength;	
}
int sgIP_TCP_Recv(sgIP_Record_TCP * rec, char * databuf, int buflength, int flags) {
	if(!rec || !databuf) return 0;
	if(rec->buf_rx_in==rec->buf_rx_out) return 0; // no data
	SGIP_INTR_PROTECT();
	int rxlen = rec->buf_rx_out - rec->buf_rx_in;
	if(rxlen<0) rxlen+=SGIP_TCP_RECEIVEBUFFERLENGTH;
	if(buflength>rxlen) buflength=rxlen;
	int i,j;
	j=rec->buf_rx_in;
	for(i=0;i<buflength;i++) {
		databuf[i]=rec->buf_rx[j++];
		if(j==SGIP_TCP_RECEIVEBUFFERLENGTH) j=0;
	}
	rec->buf_rx_in=j;

	SGIP_INTR_UNPROTECT();
	return buflength;
}
