// DSWifi Project - sgIP Internet Protocol Stack Implementation
// Copyright (C) 2005 Stephen Stair - sgstair@akkit.org - http://www.akkit.org
/****************************************************************************** 
DSWifi Lib and test materials are licenced under the MIT open source licence:
Copyright (c) 2005 Stephen Stair

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
******************************************************************************/

#include "sgIP_sockets.h"
#include "sgIP_TCP.h"
#include "sgIP_UDP.h"
#include "sgIP_ICMP.h"


sgIP_socket_data socketlist[SGIP_SOCKET_MAXSOCKETS];

void sgIP_sockets_Init() {
	int i;
	for(i=0;i<SGIP_SOCKET_MAXSOCKETS;i++) {
		socketlist[i].conn_ptr=0;
		socketlist[i].flags = 0;
	}
}


int socket(int domain, int type, int protocol) {
	int s;
	if(domain!=AF_INET) return -1;
	if(protocol!=0) return -1;
	if(type!=SOCK_DGRAM && type!=SOCK_STREAM) return -1;
	SGIP_INTR_PROTECT();
	for(s=0;s<SGIP_SOCKET_MAXSOCKETS;s++) if(!(socketlist[s].flags&SGIP_SOCKET_FLAG_ACTIVE)) break;
	if(type==SOCK_STREAM) {
		socketlist[s].flags=SGIP_SOCKET_FLAG_ACTIVE | SGIP_SOCKET_FLAG_TYPE_TCP;
		socketlist[s].conn_ptr=sgIP_TCP_AllocRecord();
	} else if(type==SOCK_DGRAM) {
		socketlist[s].flags=SGIP_SOCKET_FLAG_ACTIVE | SGIP_SOCKET_FLAG_TYPE_UDP;
		socketlist[s].conn_ptr=sgIP_UDP_AllocRecord();		
	} else {
		SGIP_INTR_UNPROTECT();
		return -1;
	}
#ifdef SGIP_SOCKET_DEFAULT_NONBLOCK
	socketlist[s].flags|=SGIP_SOCKET_FLAG_NONBLOCKING;
#endif
	SGIP_INTR_UNPROTECT();
	return s+1;
}
int close(int socket) {
	if(socket<1 || socket>SGIP_SOCKET_MAXSOCKETS) return -1;
	SGIP_INTR_PROTECT();
	socket--;
	if((socketlist[socket].flags&SGIP_SOCKET_FLAG_TYPEMASK)==SGIP_SOCKET_FLAG_TYPE_TCP) {
		sgIP_TCP_FreeRecord((sgIP_Record_TCP *)socketlist[socket].conn_ptr);
	} else if((socketlist[socket].flags&SGIP_SOCKET_FLAG_TYPEMASK)==SGIP_SOCKET_FLAG_TYPE_UDP) {
		sgIP_UDP_FreeRecord((sgIP_Record_UDP *)socketlist[socket].conn_ptr);
	}
	socketlist[socket].conn_ptr=0;
	socketlist[socket].flags=0;
	SGIP_INTR_UNPROTECT();
	return 0;
}

int bind(int socket, const struct sockaddr * addr, int addr_len) {
	if(socket<1 || socket>SGIP_SOCKET_MAXSOCKETS) return -1;
	SGIP_INTR_PROTECT();
	int retval=-1;
	socket--;
	if((socketlist[socket].flags&SGIP_SOCKET_FLAG_TYPEMASK)==SGIP_SOCKET_FLAG_TYPE_TCP) {
	} else if((socketlist[socket].flags&SGIP_SOCKET_FLAG_TYPEMASK)==SGIP_SOCKET_FLAG_TYPE_UDP) {
		retval=sgIP_UDP_Bind((sgIP_Record_UDP *)socketlist[socket].conn_ptr,((struct sockaddr_in *)addr)->sin_port,((struct sockaddr_in *)addr)->sin_addr.s_addr);
	}
	SGIP_INTR_UNPROTECT();
	return 0;
}
int connect(int socket, const struct sockaddr * addr, int addr_len);
int send(int socket, const void * data, int sendlength, int flags);
int recv(int socket, void * data, int recvlength, int flags);
int sendto(int socket, const void * data, int sendlength, int flags, const struct sockaddr * addr, int addr_len) {
	if(socket<1 || socket>SGIP_SOCKET_MAXSOCKETS) return -1;
	if(!addr) return -1;
	SGIP_INTR_PROTECT();
	int retval=-1;
	socket--;
	if((socketlist[socket].flags&SGIP_SOCKET_FLAG_TYPEMASK)==SGIP_SOCKET_FLAG_TYPE_TCP) {
	} else if((socketlist[socket].flags&SGIP_SOCKET_FLAG_TYPEMASK)==SGIP_SOCKET_FLAG_TYPE_UDP) {
		retval=sgIP_UDP_SendTo((sgIP_Record_UDP *)socketlist[socket].conn_ptr,data,sendlength,flags,((struct sockaddr_in *)addr)->sin_addr.s_addr,((struct sockaddr_in *)addr)->sin_port);
	}
	SGIP_INTR_UNPROTECT();
	return retval;
}
int recvfrom(int socket, void * data, int recvlength, int flags, struct sockaddr * addr, int * addr_len) {
	if(socket<1 || socket>SGIP_SOCKET_MAXSOCKETS) return -1;
	if(!addr) return -1;
	SGIP_INTR_PROTECT();
	int retval=-1;
	socket--;
	if((socketlist[socket].flags&SGIP_SOCKET_FLAG_TYPEMASK)==SGIP_SOCKET_FLAG_TYPE_TCP) {
	} else if((socketlist[socket].flags&SGIP_SOCKET_FLAG_TYPEMASK)==SGIP_SOCKET_FLAG_TYPE_UDP) {
		retval=sgIP_UDP_RecvFrom((sgIP_Record_UDP *)socketlist[socket].conn_ptr,data,recvlength,flags,&(((struct sockaddr_in *)addr)->sin_addr.s_addr),&(((struct sockaddr_in *)addr)->sin_port));
		*addr_len = sizeof(struct sockaddr_in);
	}
	SGIP_INTR_UNPROTECT();
	return retval;;
}
int listen(int socket, int max_connections);
int accept(int socket, struct sockaddr * addr, int * addr_len);
int shutdown(int socket, int shutdown_type);

int ioctl(int socket, long cmd, void * arg) {
	if(socket<1 || socket>SGIP_SOCKET_MAXSOCKETS) return -1;
	socket--;
	int retval;
	retval=0;
	SGIP_INTR_PROTECT();
	switch(cmd) {
	case FIONBIO:
		if(!arg){
			retval=-1;
		} else {
			socketlist[socket].flags &= ~SGIP_SOCKET_FLAG_NONBLOCKING;
			if(*((unsigned long *)arg)) socketlist[socket].flags |= SGIP_SOCKET_FLAG_NONBLOCKING;
		}
		break;

	}
	SGIP_INTR_UNPROTECT();
	return retval;
}

int setsockopt(int socket, int level, int option_name, const void * data, int data_len) {

}
int getsockopt(int socket, int level, int option_name, void * data, int * data_len) {

}
