// DS Wifi interface code
// Copyright (C) 2005 Stephen Stair - sgstair@akkit.org - http://www.akkit.org
// wifi_arm9.c - arm9 wifi support code
/****************************************************************************** 
DSWifi Lib and test materials are licenced under the MIT open source licence:
Copyright (c) 2005 Stephen Stair

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
******************************************************************************/


#include <NDS/NDS.h>
#include "../../arm7/source/dsregs.h"

#include "wifi_arm9.h"
#include <stdarg.h>

#ifdef WIFI_USE_TCP_SGIP

#include "sgIP.h"


sgIP_Hub_HWInterface * wifi_hw;


int sgIP_DisableInterrupts() {
	int a;
	a=IME;
	IME=0;
	return a;
}
void sgIP_RestoreInterrupts(int old_ime) {
	IME=old_ime;
}

#endif





void ethhdr_print(char f, void * d) {
	char buffer[33];
	int i;
	int t,c;
	buffer[0]=f;
	buffer[1]=':';
	buffer[14]=' ';
	buffer[27]=' ';
	buffer[32]=0;
	for(i=0;i<6;i++) {
		t=((u8 *)d)[i];
		c=t&15;
		if(c>9) c+='A'-10; else c+='0';
		buffer[3+i*2]=c;
		c=(t>>4)&15;
		if(c>9) c+='A'-10; else c+='0';
		buffer[2+i*2]=c;

		t=((u8 *)d)[i+6];
		c=t&15;
		if(c>9) c+='A'-10; else c+='0';
		buffer[16+i*2]=c;
		c=(t>>4)&15;
		if(c>9) c+='A'-10; else c+='0';
		buffer[15+i*2]=c;
	}
	for(i=0;i<2;i++) {
		t=((u8 *)d)[i+12];
		c=t&15;
		if(c>9) c+='A'-10; else c+='0';
		buffer[29+i*2]=c;
		c=(t>>4)&15;
		if(c>9) c+='A'-10; else c+='0';
		buffer[28+i*2]=c;
	}
	SGIP_DEBUG_MESSAGE((buffer));
}



#include "..\..\arm7\source\spinlock.h" // .h file with code for spinlocking in it.

Wifi_MainStruct Wifi_Data_Struct;

volatile Wifi_MainStruct * WifiData = 0;

WifiPacketHandler packethandler = 0;


void erasemem(void * mem, int length) {
	int i;
	char * m = (char *)mem;
	for(i=0;i<length;i++)
		m[i]=0;
}

void Wifi_CopyMacAddr(volatile void * dest, volatile void * src) {
	((u16 *)dest)[0]=((u16 *)src)[0];
	((u16 *)dest)[1]=((u16 *)src)[1];
	((u16 *)dest)[2]=((u16 *)src)[2];
}

int Wifi_CmpMacAddr(volatile void * mac1,volatile  void * mac2) {
	return (((u16 *)mac1)[0]==((u16 *)mac2)[0]) && (((u16 *)mac1)[1]==((u16 *)mac2)[1]) && (((u16 *)mac1)[2]==((u16 *)mac2)[2]);
}



u32 Wifi_TxBufferWordsAvailable() {
	s32 size=WifiData->txbufIn-WifiData->txbufOut-1;
	if(size<0) size += WIFI_TXBUFFER_SIZE/2;
	return size;
}
void Wifi_TxBufferWrite(s32 start, s32 len, u16 * data) {
	int writelen;
	while(len>0) {
		writelen=len;
		if(writelen>(WIFI_TXBUFFER_SIZE/2)-start) writelen=(WIFI_TXBUFFER_SIZE/2)-start;
		len-=writelen;
		while(writelen) {
			WifiData->txbufData[start++]=*(data++);
			writelen--;
		}
		start=0;
	}
}

int Wifi_RxRawReadPacket(s32 packetID, s32 readlength, u16 * data) {
	int readlen,read_data;
	readlength= (readlength+1)/2;
	read_data=0;
	while(readlength>0) {
		readlen=readlength;
		if(readlen>(WIFI_RXBUFFER_SIZE/2)-packetID) readlen=(WIFI_RXBUFFER_SIZE/2)-packetID;
		readlength-=readlen;
		read_data+=readlen;
		while(readlen>0) {
			*(data++) = WifiData->rxbufData[packetID++];
			readlen--;
		}
		packetID=0;
	}
	return read_data;
}

u16 Wifi_RxReadOffset(s32 base, s32 offset) {
	base+=offset;
	if(base>=(WIFI_RXBUFFER_SIZE/2)) base -= (WIFI_RXBUFFER_SIZE/2);
	return WifiData->rxbufData[base];
}

// datalen = size of packet from beginning of 802.11 header to end, but not including CRC.
int Wifi_RawTxFrame(u16 datalen, u16 rate, u16 * data) {
	Wifi_TxHeader txh;
	int sizeneeded;
	int base;
	sizeneeded=((datalen+12+4+3)/4)*2;
	if(sizeneeded>Wifi_TxBufferWordsAvailable()) {WifiData->stats[WSTAT_TXQUEUEDREJECTED]++; return -1; }
	txh.tx_rate=rate;
	txh.tx_length=datalen+4;
	base = WifiData->txbufOut;
	Wifi_TxBufferWrite(base,6,(u16 *)&txh);
	base += 6;
	if(base>=(WIFI_TXBUFFER_SIZE/2)) base -= WIFI_TXBUFFER_SIZE/2;
	Wifi_TxBufferWrite(base,((datalen+3)/4)*2,data);
	base += ((datalen+3)/4)*2;
	if(base>=(WIFI_TXBUFFER_SIZE/2)) base -= WIFI_TXBUFFER_SIZE/2;
	WifiData->txbufOut=base;
	WifiData->stats[WSTAT_TXQUEUEDPACKETS]++;
	WifiData->stats[WSTAT_TXQUEUEDBYTES]+=sizeneeded;
}


void Wifi_RawSetPacketHandler(WifiPacketHandler wphfunc) {
	packethandler=wphfunc;
}
void Wifi_DisableWifi() {
	WifiData->reqMode=WIFIMODE_DISABLED;
	WifiData->reqReqFlags &= ~WFLAG_REQ_APCONNECT;
}
void Wifi_EnableWifi() {
	WifiData->reqMode=WIFIMODE_NORMAL;
	WifiData->reqReqFlags &= ~WFLAG_REQ_APCONNECT;
}
void Wifi_ScanMode() {
	WifiData->reqMode=WIFIMODE_SCAN;
	WifiData->reqReqFlags &= ~WFLAG_REQ_APCONNECT;
}
int Wifi_GetNumAP() {
	int i,j;
	j=0;
	for(i=0;i<WIFI_MAX_AP;i++) if(WifiData->aplist[i].flags&WFLAG_APDATA_ACTIVE) j++;
	return j;
}
int Wifi_GetAPData(int apnum, Wifi_AccessPoint * apdata) {
	int i,j;
	if(!apdata) return WIFI_RETURN_PARAMERROR;
	j=0;
	for(i=0;i<WIFI_MAX_AP;i++){
		if(j==apnum) {
			if(Spinlock_Acquire(WifiData->aplist[i])==SPINLOCK_OK) {
				// additionally calculate average RSSI here
				WifiData->aplist[i].rssi=0;
				for(j=0;j<8;j++) {
					WifiData->aplist[i].rssi+=WifiData->aplist[i].rssi_past[j];
				}
				WifiData->aplist[i].rssi = WifiData->aplist[i].rssi >> 3;
				*apdata = WifiData->aplist[i]; // yay for struct copy!
				Spinlock_Release(WifiData->aplist[i]);
				return WIFI_RETURN_OK;
			} else {
				return WIFI_RETURN_LOCKFAILED;
			}
		}
		if(WifiData->aplist[i].flags&WFLAG_APDATA_ACTIVE) j++;
	}
	return WIFI_RETURN_ERROR;
}

int Wifi_ConnectAP(Wifi_AccessPoint * apdata, int wepmode, int wepkeyid, u8 * wepkey) {
	int i;
	Wifi_CopyMacAddr(WifiData->bssid9, apdata->bssid);
	Wifi_CopyMacAddr(WifiData->apmac9, apdata->bssid);
	WifiData->ssid9[0]=apdata->ssid_len;
	for(i=0;i<32;i++) {
		WifiData->ssid9[i+1]=apdata->ssid[i];
	}
	WifiData->apchannel9=apdata->channel;
	WifiData->wepmode9=wepmode;
	WifiData->wepkeyid9=wepkeyid;
	for(i=0;i<20;i++) {
		WifiData->wepkey9[i]=wepkey[i];
	}
	for(i=0;i<16;i++) WifiData->baserates9[i]=apdata->base_rates[i];
	WifiData->reqMode=WIFIMODE_NORMAL;
	WifiData->reqReqFlags |= WFLAG_REQ_APCONNECT | WFLAG_REQ_APCOPYVALUES;
}
int Wifi_AssocStatus() {
	switch(WifiData->curMode) {
	case WIFIMODE_DISABLED:
	case WIFIMODE_NORMAL:
	case WIFIMODE_SCAN:
	case WIFIMODE_PROMISCUOUS:
	case WIFIMODE_DISASSOCIATE:
		return ASSOCSTATUS_DISCONNECTED;
	case WIFIMODE_ASSOCIATE:
		switch(WifiData->authlevel) {
		case WIFI_AUTHLEVEL_DISCONNECTED:
			return ASSOCSTATUS_AUTHENTICATING;
		case WIFI_AUTHLEVEL_AUTHENTICATED:
		case WIFI_AUTHLEVEL_DEASSOCIATED:
			return ASSOCSTATUS_ASSOCIATING;
		case WIFI_AUTHLEVEL_ASSOCIATED:
			return ASSOCSTATUS_ASSOCIATED;
		}
		break;
	case WIFIMODE_ASSOCIATED:
		return ASSOCSTATUS_ASSOCIATED;
	case WIFIMODE_CANNOTASSOCIATE:
		return ASSOCSTATUS_CANNOTCONNECT;
	}
	return ASSOCSTATUS_DISCONNECTED;
}
int Wifi_DisconnectAP() {
	WifiData->reqMode=WIFIMODE_NORMAL;
	WifiData->reqReqFlags &= ~WFLAG_REQ_APCONNECT;
}


#ifdef WIFI_USE_TCP_SGIP



int Wifi_TransmitFunction(sgIP_Hub_HWInterface * hw, sgIP_memblock * mb) {
	// convert ethernet frame into wireless frame and output.
	// ethernet header: 6byte dest, 6byte src, 2byte protocol_id
	// assumes individual pbuf len is >=14 bytes, it's pretty likely ;) - also hopes pbuf len is a multiple of 2 :|
	int base,framelen, hdrlen, writelen;
   int copytotal, copyexpect;
	u16 framehdr[6+12+2];
	sgIP_memblock * t;
	framelen=mb->totallength-14+8;

	if(framelen+40>Wifi_TxBufferWordsAvailable()*2) { // error, can't send this much!
		SGIP_DEBUG_MESSAGE(("Transmit:err_space"));
      sgIP_memblock_free(mb);
		return 0; //?
	}
	
	ethhdr_print('T',mb->datastart);
	framehdr[0]=0;
	framehdr[1]=0;
	framehdr[2]=0;
	framehdr[3]=0;
	framehdr[4]=0; // rate, will be filled in by the arm7.
	hdrlen=18;
	framehdr[7]=0;

	if(WifiData->curReqFlags&WFLAG_REQ_APADHOC) { // adhoc mode
		framehdr[6]=0x0008;
		Wifi_CopyMacAddr(framehdr+14,WifiData->bssid7);
		Wifi_CopyMacAddr(framehdr+11,WifiData->MacAddr);
		Wifi_CopyMacAddr(framehdr+8,((u8 *)mb->datastart));
	} else {
		framehdr[6]=0x0108;
		Wifi_CopyMacAddr(framehdr+8,WifiData->bssid7);
		Wifi_CopyMacAddr(framehdr+11,WifiData->MacAddr);
		Wifi_CopyMacAddr(framehdr+14,((u8 *)mb->datastart));
	}
	if(WifiData->wepmode7)  { framehdr[6] |=0x4000; hdrlen=20; }
	framehdr[17] = 0;
	framehdr[18] = 0; // wep IV, will be filled in if needed on the arm7 side.
	framehdr[19] = 0;

   framehdr[5]=framelen+hdrlen*2-12+4;
   copyexpect= ((framelen+hdrlen*2-12+4) +12 -4 +1)/2;
   copytotal=0;

	WifiData->stats[WSTAT_TXQUEUEDPACKETS]++;
	WifiData->stats[WSTAT_TXQUEUEDBYTES]+=framelen+hdrlen*2;

	base = WifiData->txbufOut;
	Wifi_TxBufferWrite(base,hdrlen,framehdr);
	base += hdrlen;
   copytotal+=hdrlen;
	if(base>=(WIFI_TXBUFFER_SIZE/2)) base -= WIFI_TXBUFFER_SIZE/2;

	// add LLC header
	framehdr[0]=0xAAAA;
	framehdr[1]=0x0003;
	framehdr[2]=0x0000;
	framehdr[3]=((u16 *)mb->datastart)[6]; // frame type

	Wifi_TxBufferWrite(base,4,framehdr);
	base += 4;
   copytotal+=4;
	if(base>=(WIFI_TXBUFFER_SIZE/2)) base -= WIFI_TXBUFFER_SIZE/2;

	t=mb;
	writelen=(mb->thislength-14);
	if(writelen) {
		Wifi_TxBufferWrite(base,(writelen+1)/2,((u16 *)mb->datastart)+7);
		base+=(writelen+1)/2;
      copytotal+=(writelen+1)/2;
		if(base>=(WIFI_TXBUFFER_SIZE/2)) base -= WIFI_TXBUFFER_SIZE/2;
	}
	while(mb->next) {
		mb=mb->next;
		writelen=mb->thislength;
		Wifi_TxBufferWrite(base,(writelen+1)/2,((u16 *)mb->datastart));
		base+=(writelen+1)/2;
      copytotal+=(writelen+1)/2;
		if(base>=(WIFI_TXBUFFER_SIZE/2)) base -= WIFI_TXBUFFER_SIZE/2;
	}

	WifiData->txbufOut=base; // update fifo out pos, done sending packet.

	sgIP_memblock_free(t); // free packet, as we're the last stop on this chain.

   if(copytotal!=copyexpect) {
      SGIP_DEBUG_MESSAGE(("Tx exp:%i que:%i",copyexpect,copytotal));
   }
   SGIP_DEBUG_MESSAGE(("sp:%i d:%04X",WifiData->stats[WSTAT_TXPACKETS],WifiData->stats[WSTAT_DEBUG]));

	return 0;
}

int Wifi_Interface_Init(sgIP_Hub_HWInterface * hw) {
	hw->MTU=2300;
	hw->ipaddr=(192)|(168<<8)|(1<<16)|(151<<24);
	hw->snmask=0x00FFFFFF;
	hw->gateway=(192)|(168<<8)|(1<<16)|(1<<24);
	hw->dns[0]=(192)|(168<<8)|(1<<16)|(1<<24);
	hw->hwaddrlen=6;
	Wifi_CopyMacAddr(hw->hwaddr,WifiData->MacAddr);
	hw->userdata=0;
}

u16 count_100ms;

void Wifi_Timer_50ms() {
	count_100ms++;
	if(count_100ms>=2) {
		count_100ms=0;
		sgIP_ARP_Timer100ms();
	}
	Wifi_Update();
}

#endif

unsigned long Wifi_Init(int initflags) {
	erasemem(&Wifi_Data_Struct,sizeof(Wifi_Data_Struct));
	WifiData = (Wifi_MainStruct *) (((u32)&Wifi_Data_Struct)| 0x00400000); // should prevent the cache from eating us alive.

#ifdef WIFI_USE_TCP_SGIP

	sgIP_Init();

#endif

	WifiData->flags9 = WFLAG_ARM9_ACTIVE;
	return (u32) &Wifi_Data_Struct;
}

int Wifi_CheckArm7Init() {
	if(!WifiData) return 0;
	return (WifiData->flags7 & WFLAG_ARM7_ACTIVE);
}


void Wifi_Update() {
	int cnt;
	int base, base2, len, fulllen;
	if(!WifiData) return;

#ifdef WIFI_USE_TCP_SGIP

	if(!(WifiData->flags9&WFLAG_ARM9_ARM7READY)) {
		if(WifiData->flags7 & WFLAG_ARM7_ACTIVE) {
			WifiData->flags9 |=WFLAG_ARM9_ARM7READY;
			// add network interface.
			wifi_hw = sgIP_Hub_AddHardwareInterface(&Wifi_TransmitFunction,&Wifi_Interface_Init);
		}
	}
	if(WifiData->authlevel!=WIFI_AUTHLEVEL_ASSOCIATED && WifiData->flags9&WFLAG_ARM9_NETUP) {
	//	netif_set_down(&lwip_netif);
		WifiData->flags9 &= ~(WFLAG_ARM9_NETUP);
	} else if(WifiData->authlevel==WIFI_AUTHLEVEL_ASSOCIATED && !(WifiData->flags9&WFLAG_ARM9_NETUP)) {
	//	netif_set_up(&lwip_netif);
		WifiData->flags9 |= (WFLAG_ARM9_NETUP);
		if(WifiData->iptype==WIFI_IPTYPE_DHCP) {
	//		printtop("dhcp_start");
	//		dhcp_start( &lwip_netif );
		} else {
			//printtop("dhcp_inform");
			//dhcp_inform( &lwip_netif );
	//		etharp_request( &lwip_netif, &(lwip_netif.ip_addr)); // send grat arp.
		}
	}

#endif

	// check for received packets, forward to whatever wants them.
	cnt=0;
	while(WifiData->rxbufIn!=WifiData->rxbufOut) {
		base = WifiData->rxbufIn;
		len=Wifi_RxReadOffset(base,4);
		fulllen=((len+3)&(~3))+12;
#ifdef WIFI_USE_TCP_SGIP
		// Do lwIP interfacing for rx here
		if((Wifi_RxReadOffset(base,6)&0x01CF)==0x0008) // if it is a non-null data packet coming from the AP (toDS==0)
		{
			u16 framehdr[6+12+2+4];
			sgIP_memblock * mb;
			int hdrlen;
			base2=base;
			Wifi_RxRawReadPacket(base,22*2,framehdr);

			if((framehdr[8]==((u16 *)WifiData->MacAddr)[0] && framehdr[9]==((u16 *)WifiData->MacAddr)[1] && framehdr[10]==((u16 *)WifiData->MacAddr)[2]) ||
				(framehdr[8]==0xFFFF && framehdr[9]==0xFFFF && framehdr[10]==0xFFFF)) {
				// destination matches our mac address, or the broadcast address.
				if(framehdr[6]&0x4000) { // wep enabled
					base2+=24; hdrlen=28;  // base2+=[wifi hdr 12byte]+[802 header hdrlen]+[slip hdr 8byte]
				} else { 
					base2+=22; hdrlen=24;
				}
				// check for LLC/SLIP header...
				if(Wifi_RxReadOffset(base2-4,0)==0xAAAA && Wifi_RxReadOffset(base2-4,1)==0x0003 && Wifi_RxReadOffset(base2-4,2)==0) {
					mb = sgIP_memblock_allocHW(14,len-8-hdrlen);
					if(mb) {
						if(base2>=(WIFI_RXBUFFER_SIZE/2)) base2-=(WIFI_RXBUFFER_SIZE/2);
						Wifi_RxRawReadPacket(base2,(len-8-hdrlen)&(~1),((u16 *)mb->datastart)+7);
						if(len&1) ((u8 *)mb->datastart)[len+14-1-8-hdrlen]=Wifi_RxReadOffset(base2,((len-8-hdrlen)/2))&255;
						Wifi_CopyMacAddr(mb->datastart,framehdr+8); // copy dest
						if(Wifi_RxReadOffset(base,6)&0x0200) { // from DS set?
							Wifi_CopyMacAddr(((u8 *)mb->datastart)+6,framehdr+14); // copy src from adrs3
						} else {
							Wifi_CopyMacAddr(((u8 *)mb->datastart)+6,framehdr+11); // copy src from adrs2
						}
						((u16 *)mb->datastart)[6]=framehdr[(hdrlen/2)+6+3]; // assume LLC exists and is 8 bytes.

						ethhdr_print('R',mb->datastart);

						// Done generating recieved data packet... now distribute it.
						sgIP_Hub_ReceiveHardwarePacket(wifi_hw,mb);

					}
				}
			}
		}

#endif

		// check if we have a handler
		if(packethandler) {
			base2=base+6;
			if(base2>=(WIFI_RXBUFFER_SIZE/2)) base2-=(WIFI_RXBUFFER_SIZE/2);
			(*packethandler)(base2,len);
		}

		base+=fulllen/2;
		if(base>=(WIFI_RXBUFFER_SIZE/2)) base-=(WIFI_RXBUFFER_SIZE/2);
		WifiData->rxbufIn=base;

		if(cnt++>80) break;
	}
}


//////////////////////////////////////////////////////////////////////////
// Ip addr get/set functions
#ifdef WIFI_USE_TCP_SGIP

u32 Wifi_GetIP() {
	if(wifi_hw) return wifi_hw->ipaddr;
	return 0;
}

void Wifi_SetStaticIP(u32 IPaddr, u32 gateway, u32 subnetmask, u32 dns1, u32 dns2) {
	if(wifi_hw) {
		wifi_hw->ipaddr=IPaddr;
		wifi_hw->gateway=gateway;
		wifi_hw->snmask=subnetmask;
		wifi_hw->dns[0]=dns1;
		wifi_hw->dns[1]=dns2;
		// reset arp cache...
		sgIP_ARP_FlushInterface(wifi_hw);
	}
}

void Wifi_SetDHCP() {


}

#endif
