// DSWifi Test Project - copyright Stephen Stair 2005
// Based on template project in ndslib by wntrmute
/****************************************************************************** 
DSWifi Lib and test materials are licenced under the MIT open source licence:
Copyright (c) 2005 Stephen Stair

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
******************************************************************************/

#include "nds.h"
#include <nds/arm9/console.h> //basic print funcionality
#include <stdio.h>
#include <Stdlib.h>
#include <string.h>
#include <stdarg.h>

#include "dswifi9.h"
#include "sys\socket.h"
#include "netinet\in.h"

#include "ascii_bin.h"

#define		VRAM1		((u16 *) 0x06000000)
#define		VRAM2		((u16 *) 0x06200000)
#define		VCOUNT		(*((u16 volatile *) 0x04000006))
#define PAL_BG1	((volatile u16 *)0x05000000)
#define PAL_BG2	((volatile u16 *)0x05000400)
#define RGB(r,g,b) ((r) | ((g)<<5) | ((b)<<10))

touchPosition touchXY;

Wifi_AccessPoint global_connectAP;
unsigned char global_wepkeys[4][32];
int global_wepkeyid, global_wepmode;
int global_dhcp; // 0=none, 1=get IP, 2=get DNS
unsigned long global_ipaddr, global_snmask, global_gateway, global_dns1,global_dns2;

int psych_mode;
int framectr;
const u16 psychtab1[8] = {
	0x7FFF, 0x7F7B, 0x7EF7, 0x7E73, 0x7E10, 0x7E73, 0x7EF7, 0x7F7B
};
const u16 psychtab2[8] = {
	0x1F, 0x2FF, 0x3EF, 0x1FE0, 0x7FE0, 0x7CE0, 0x7C0F, 0x5C1F
};
const u16 psychtab3[8] = {
	0x1F, 0x173, 0x2E7, 0xF60, 0x3DE0, 0x6C60, 0x5C07, 0x2C13
};

void * sgIP_malloc(int size) { return malloc(size); }
void sgIP_free(void * ptr) { free(ptr); }



void WaitVbl() {
	int i;
	while(VCOUNT>192);
	while(VCOUNT<192);
	scanKeys();
	framectr++;
	touchXY=touchReadXY();
	switch(psych_mode) {
	case 0:  // nothing
	case 1: // normal
		for(i=0;i<8;i++) {
			PAL_BG1[i+1]=RGB(31,31,31);
			PAL_BG2[i+1]=RGB(31,31,31);
		}
		break;
	case 2: // psych1
		for(i=0;i<8;i++) {
			PAL_BG1[i+1]=psychtab1[((framectr>>2)+i)&7];
			PAL_BG2[i+1]=psychtab1[((framectr>>2)-i)&7];
		}
		break;	
	case 13: // psych2
		for(i=0;i<8;i++) {
			PAL_BG1[i+1]=psychtab2[((framectr>>2)+i)&7];
			PAL_BG2[i+1]=psychtab2[((framectr>>2)-i)&7];
		}
		break;	
	case 14: // psych3
		for(i=0;i<8;i++) {
			PAL_BG1[i+1]=psychtab3[((framectr>>2)+i)&7];
			PAL_BG2[i+1]=psychtab3[((framectr>>2)-i)&7];
		}
		break;
	case 3: // loop around
		psych_mode=1;
		break;
	default:
		psych_mode=0;
	}
}

 // 2bit top, 2bit right, 2bit bottom, 2bit left (0bTTRRBBLL)
unsigned char pipe_dream[192]; // auto generated;
const unsigned char reverse_pipe_dream[256] = { //(0bTTRRBBLL)
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 00
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 10
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 20
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 30
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 40
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 50
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 60
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 70
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 80
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // 90
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // A0
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0x44, 0x45, 0x46, 0x89, 0x09, // B0
	0x06, 0x8A, 0x88, 0x0A, 0x82, 0x81, 0x42, 0x05, 
	0x50, 0x51, 0x15, 0x54, 0x11, 0x55, 0x64, 0x98, // C0
	0xA0, 0x28, 0xA2, 0x2A, 0xA8, 0x22, 0xAA, 0x62, 
	0x91, 0x26, 0x19, 0x90, 0x60, 0x24, 0x18, 0x99, // D0
	0x66, 0x41, 0x14, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // E0
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, // F0
	0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 
};

void ScrollUpTop() {
	int i;
	for(i=0;i<23*32;i++)
		VRAM1[0x07C00+i]=VRAM1[0x07C00+i+32];
	for(i=0;i<32;i++) {
		VRAM1[0x07C00+23*32+i]=' ';
	}
}
void ClsBtm() {
	int i;
	for(i=0;i<24*32;i++)
		VRAM2[0x07C00+i]=0x20;
}
void printtop(const char * str) {
	int x;
	ScrollUpTop();
	x=0;
	while(*str) {
		if(x==32) {x=0; ScrollUpTop();}
		VRAM1[0x07C00+23*32+x++]=*str;
		str++;
	}
}

void sgIP_dbgprint(char * txt, ...) {
	char buffer[256];
	va_list args;
	va_start(args,txt);
	vsprintf(buffer,txt,args);
	printtop(buffer);		
}
void printbtm(int x, int y, const char * str) {
	while(*str) {
		if(x==32) {x=0; y++;}
		if(y==24) return;
		VRAM2[0x07C00+y*32+x++]=*str;
		str++;
	}
}
void printbtmn(int x, int y, const char * str, int n) {
	while(*str && 0<=--n) {
		if(x==32) {x=0; y++;}
		if(y==24) return;
		VRAM2[0x07C00+y*32+x++]=*str;
		str++;
	}
}
void gen_pipe_dream() {
	int i;
	for(i=0;i<192;i++) {
		pipe_dream[i]=0;
	}
	for(i=0;i<256;i++) {
		if(reverse_pipe_dream[i]!=0xFF) {
			if(pipe_dream[reverse_pipe_dream[i]]) {
				printtop("Error: Duplicate pipe!");
			} else {
				pipe_dream[reverse_pipe_dream[i]] = i;
			}
		}
	}
}
void btm_drawch_pipedream(int x_, int y_, int c_) {
	unsigned char pdsrc, pddst;
	pdsrc=reverse_pipe_dream[c_];
	pddst=reverse_pipe_dream[VRAM2[0x07C00+y_*32+x_]];
	if(pdsrc!=0xFF && pddst!=0xFF) {
		if(!(pdsrc & 0xC0)) pdsrc |= (pddst&0xC0);
		if(!(pdsrc & 0x30)) pdsrc |= (pddst&0x30);
		if(!(pdsrc & 0xC)) pdsrc |= (pddst&0xC);
		if(!(pdsrc & 0x3)) pdsrc |= (pddst&0x3);
		pddst= pipe_dream[pdsrc];
		if(pddst!=0) c_ = pddst;
	}
	VRAM2[0x07C00+y_*32+x_]=c_;
}
void inline btm_drawch(int x_, int y_, int c_) {
	VRAM2[0x07C00+y_*32+x_]=c_;
}
void btm_drawrect1(int x1,int y1, int x2, int y2) {
	int i;
	if(x1<0 || y1<0 || x2>31 || y2>23) return;
	for(i=x1+1;i<x2;i++) { btm_drawch(i,y1,0xC4); btm_drawch(i,y2,0xC4); }
	for(i=y1+1;i<y2;i++) { btm_drawch(x1,i,0xB3); btm_drawch(x2,i,0xB3); }
	btm_drawch(x1,y1,0xDA);btm_drawch(x1,y2,0xC0);btm_drawch(x2,y1,0xBF);btm_drawch(x2,y2,0xD9);
}
void btm_drawrect2(int x1,int y1, int x2, int y2) {
	int i;
	if(x1<0 || y1<0 || x2>31 || y2>23) return;
	for(i=x1+1;i<x2;i++) { btm_drawch(i,y1,0xCD); btm_drawch(i,y2,0xCD); }
	for(i=y1+1;i<y2;i++) { btm_drawch(x1,i,0xBA); btm_drawch(x2,i,0xBA); }
	btm_drawch(x1,y1,0xC9);btm_drawch(x1,y2,0xC8);btm_drawch(x2,y1,0xBB);btm_drawch(x2,y2,0xBC);
}
void btm_drawbutton(int x1,int y1, int x2, int y2, int s, const char * btntxt) {
	if(s) btm_drawrect2(x1,y1,x2,y2); else btm_drawrect1(x1,y1,x2,y2); 
	s=strlen(btntxt);
	if(s>(x2-x1-1)) {
		printbtmn(x1+1,(y1+y2)>>1,btntxt,(x2-x1-1));
	} else {
		printbtm((x1+x2-s+1)>>1,(y1+y2)>>1,btntxt);
	}
}

void btm_pddrawrect1(int x1,int y1, int x2, int y2) {
	int i;
	if(x1<0 || y1<0 || x2>31 || y2>23) return;
	for(i=x1+1;i<x2;i++) { btm_drawch_pipedream(i,y1,0xC4); btm_drawch_pipedream(i,y2,0xC4); }
	for(i=y1+1;i<y2;i++) { btm_drawch_pipedream(x1,i,0xB3); btm_drawch_pipedream(x2,i,0xB3); }
	btm_drawch_pipedream(x1,y1,0xDA);btm_drawch_pipedream(x1,y2,0xC0);btm_drawch_pipedream(x2,y1,0xBF);btm_drawch_pipedream(x2,y2,0xD9);
}
void btm_pddrawrect2(int x1,int y1, int x2, int y2) {
	int i;
	if(x1<0 || y1<0 || x2>31 || y2>23) return;
	for(i=x1+1;i<x2;i++) { btm_drawch_pipedream(i,y1,0xCD); btm_drawch_pipedream(i,y2,0xCD); }
	for(i=y1+1;i<y2;i++) { btm_drawch_pipedream(x1,i,0xBA); btm_drawch_pipedream(x2,i,0xBA); }
	btm_drawch_pipedream(x1,y1,0xC9);btm_drawch_pipedream(x1,y2,0xC8);btm_drawch_pipedream(x2,y1,0xBB);btm_drawch_pipedream(x2,y2,0xBC);
}
void btm_pddrawbutton(int x1,int y1, int x2, int y2, int s, const char * btntxt) {
	if(s) btm_pddrawrect2(x1,y1,x2,y2); else btm_pddrawrect1(x1,y1,x2,y2); 
	s=strlen(btntxt);
	if(s>(x2-x1-1)) {
		printbtmn(x1+1,(y1+y2)>>1,btntxt,(x2-x1-1));
	} else {
		printbtm((x1+x2-s+1)>>1,(y1+y2)>>1,btntxt);
	}
}
typedef struct ap_record {
	char ssid[34];
	unsigned char mac[6];
	unsigned char channel;
	unsigned char rssi;
	unsigned short flags;
	struct ap_record * next;
} APRecord;

int Do_ConfigureWifi();
void Do_Play_Wardriving();
void Do_Play_KeyboardTest();
void Do_Play_UDPTest();


#define NUM_MAIN_MENU_ITEMS	3
const char * main_mnuitems[] = {
	//"**------------------------**"
		"Play Wardriving",
		"Play Keyboard Test",
		"Connect to an AP",
};
#define NUM_SUB_MENU_ITEMS	2
const char * sub_mnuitems[] = {
	//"**------------------------**"
		"UDP Test",
			"Return to Main Menu",
		"TCP Connect Test",
		"TCP Listen Test",
};
#define NUM_CONF_MENU_ITEMS	3
const char * conf_mnuitems[] = {
	//"**------------------------**"
		"Search for an AP",
		"Manual Configure",
		"Cancel",

		"Use Nintendo WFC Data",
};

//---------------------------------------------------------------------------------
int main(void) {
//---------------------------------------------------------------------------------
	int sekrit_counter;
   IPC->mailData=0;
   IPC->mailSize=0;
	powerON(POWER_ALL_2D | POWER_SWAP_LCDS); 
	psych_mode=0;
	u32 Wifi_pass= Wifi_Init(0);
	IPC->mailData=2;
   while(IPC->mailSize!=2)	IPC->mailData=2;
	IPC->mailData=Wifi_pass;
	*((volatile u16 *)0x0400010E) = 0;
	irqInit();
	irqSet(IRQ_TIMER3, Wifi_Timer_50ms);
	irqEnable(IRQ_TIMER3);
	*((volatile u16 *)0x0400010C) = -13106; // 13107.2 * 256 cycles = ~50ms;
	*((volatile u16 *)0x0400010E) = 0x00C2; // enable, irq, 1/256 clock

	int i;
	DISPLAY_CR=0x00012100;
	SUB_DISPLAY_CR=0x00012100;

	vramSetMainBanks(VRAM_A_MAIN_BG,VRAM_B_MAIN_BG,VRAM_C_SUB_BG,VRAM_D_LCD);
	keysInit();

	// clear vram
	for(i=0;i<32768;i++) {
		VRAM1[i]=VRAM2[i]=0;
	}
	BG0_CR=0x1f80;
	SUB_BG0_CR=0x1f80;										   

	PAL_BG1[0]=RGB(0,0,0);
	PAL_BG2[0]=RGB(0,0,0);
	WIN_IN=0x000f;
	WIN_OUT=0x00ff;
	BLEND_CR=0x00E0;
	BLEND_Y=15;
	SUB_WIN_IN=0x000f;
	SUB_WIN_OUT=0x00ff;
	SUB_BLEND_CR=0x00E0;
	SUB_BLEND_Y=15;


	//	WAIT_CR &= ~0x0880;
	for(i=0;i<256*8*4;i++) {
		((u16 *)BG_TILE_RAM(0))[i]=((u16 *)ASCII_bin)[i];
		((u16 *)BG_TILE_RAM_SUB(0))[i]=((u16 *)ASCII_bin)[i];
	}
	for(i=0;i<8;i++) {
		PAL_BG1[i+1]=RGB(31,31,31);
		PAL_BG2[i+1]=RGB(31,31,31);
	}

// set defaults
	global_connectAP.ssid_len=7;
	strcpy(global_connectAP.ssid,"linksys");
	global_connectAP.rssi=0;
	global_connectAP.channel=1;
	global_connectAP.flags=0;
	global_wepkeyid=0;
	global_wepmode=0;
	global_dhcp=0;
	global_ipaddr=0xBD01A8C0;
	global_snmask=0x00FFFFFF;
	global_gateway=0x0101A8C0;
	global_dns1=0x0101A8C0;
	global_dns2=0;


	printtop("Wifi_lib_test - Stephen Stair");
	printtop("Version 0.2b");
	printtop("Build Date:" __DATE__ " " __TIME__);
	printtop("");
	printtop("Waiting for ARM7 to init..");
	while(Wifi_CheckArm7Init()==0) WaitVbl();
	printtop("ARM7 Init Confirmed.");

	gen_pipe_dream();

	int state, touchstate;
	state=0;
	touchstate=-1;
	psych_mode=1;
	while(1) {
		sekrit_counter=0;
		state=0;touchstate=-1;
		while(1) {
			WaitVbl();
			ClsBtm();
			if(keysUp()&KEY_TOUCH && touchstate==state) { // just released
				break; // state confirmed.
			}
			touchstate=-1;
			if(keysHeld()&KEY_TOUCH) { // update touchstate
				sekrit_counter++;
				if(sekrit_counter>240) { // 4 seconds
					sekrit_counter=0;
					psych_mode++;
				}
				if(touchXY.px>16 && touchXY.px<240) {
					for(i=0;i<NUM_MAIN_MENU_ITEMS;i++) {
						if(touchXY.py>16+32*i && touchXY.py<40+32*i) { touchstate=i; }
					}
				}
				if(touchstate!=-1) state=touchstate;
			} else {
				sekrit_counter=0;
			}
			if(keysDown()&KEY_UP) {
				if(state==0) state=NUM_MAIN_MENU_ITEMS;
				state--;
			}
			if(keysDown()&KEY_DOWN) {
				state++;
				if(state==NUM_MAIN_MENU_ITEMS) state=0;
			}
			if(keysDown()&(KEY_A | KEY_START)) {
				break; // state confirmed
			}
			
			for(i=0;i<NUM_MAIN_MENU_ITEMS;i++) {
				btm_drawbutton(3,1+i*4,28,3+i*4,i==state,main_mnuitems[i]);
			}
			printbtm(0,23,"[UP/DN]Move[A/Touch]Sel");

		}
		while(keysHeld()&KEY_TOUCH) WaitVbl();

		switch(state) {
			case 0: // play wardriving
				printtop("Picked: Play Wardriving");
				Wifi_ScanMode();
				Do_Play_Wardriving();
				Wifi_DisableWifi();
				printtop("Returned from Wardriving");
				break;
			case 1: // play keyboard test
				printtop("Picked: Play Keyboard Test");
				Do_Play_KeyboardTest();
				printtop("Returned from Keyboard Test");
				break;
			case 2: // connect to an AP
				printtop("Picked: Connect to an AP");
				if(Do_ConfigureWifi()!=0) {
					printtop("Error configuring Wifi");
					break;
				}
				while(1) { // sub loop
					printtop("Entering Sub-test menu");
					state=0;touchstate=-1;
					while(1) { // sub-men
						WaitVbl();
						ClsBtm();
						if(keysUp()&KEY_TOUCH && touchstate==state) { // just released
							break; // state confirmed.
						}
						touchstate=-1;
						if(keysHeld()&KEY_TOUCH) { // update touchstate
							if(touchXY.px>16 && touchXY.px<240) {
								for(i=0;i<NUM_SUB_MENU_ITEMS;i++) {
									if(touchXY.py>16+32*i && touchXY.py<40+32*i) { touchstate=i; }
								}
							}
							if(touchstate!=-1) state=touchstate;
						}
						if(keysDown()&KEY_UP) {
							if(state==0) state=NUM_SUB_MENU_ITEMS;
							state--;
						}
						if(keysDown()&KEY_DOWN) {
							state++;
							if(state==NUM_SUB_MENU_ITEMS) state=0;
						}
						if(keysDown()&(KEY_A | KEY_START)) {
							break; // state confirmed
						}
						if(keysDown()&(KEY_B)) {
							state=NUM_SUB_MENU_ITEMS-1;
							break;
						}

						for(i=0;i<NUM_SUB_MENU_ITEMS;i++) {
							btm_drawbutton(3,1+i*4,28,3+i*4,i==state,sub_mnuitems[i]);
						}
								// "0123456789ABCDEF0123456789ABCDEF"
						printbtm(0,23,"[UP/DN]Move[A/Touch]Sel[B]Return");
					}
					while(keysHeld()&KEY_TOUCH) WaitVbl();
					if(state==(NUM_SUB_MENU_ITEMS-1)) break;
					switch(state) {
					case 0:
						Do_Play_UDPTest();
						break;

					}
				}
				printtop("Leaving Sub-Menu");
				Wifi_DisconnectAP();
				Wifi_DisableWifi();
				break;
		}
		printtop("Returning to main menu...");
	}

	return 0;
}


#define MAX_AP_WARDRIVING		4096
void Do_Play_Wardriving() {
	char renderbuf[64];
	int numaps,i,j,k,add;
	int scrolltop;
	int basetouchy,basetouchscroll,basetouchnumaps;
	Wifi_AccessPoint ap;
	APRecord * aplist;
	APRecord * aphashtabl[256], *temp;
	aplist = (APRecord *)malloc(sizeof(APRecord)*MAX_AP_WARDRIVING);
	numaps=0;
	scrolltop=0;
	basetouchscroll=-2;
	basetouchnumaps=1;
	basetouchy=96;
	for(i=0;i<256;i++) aphashtabl[i]=0;
	while(1) {
		WaitVbl();
		ClsBtm();
		// process input
		if(keysDown()&(KEY_B)) {
			break; // exit.
		}
		if(keysDown()&(KEY_UP)) {
			scrolltop--;
		}
		if(keysDown()&(KEY_DOWN)) {
			scrolltop++;
		}
		if(keysDown()&(KEY_TOUCH)) {
			basetouchscroll=-2;
			if(touchXY.px>230) {
				if(touchXY.py<16) {
					scrolltop--;
				} else if(touchXY.py>176) {
					scrolltop++;
				} else {
					basetouchy=touchXY.py;
					basetouchscroll=scrolltop;
					basetouchnumaps=numaps;
					if(basetouchnumaps==0) basetouchnumaps=1;
				}
			}
		}
		if(keysHeld()&KEY_TOUCH && basetouchscroll!=-2) {
			scrolltop=basetouchscroll+(((touchXY.py-basetouchy)*basetouchnumaps)/160);
		}
		if(scrolltop>numaps-7) scrolltop=numaps-7;
		if(scrolltop<0) scrolltop=0;

		// render::scanning line
		printbtm(0,0,"Scanning:" ); // 10
		if(numaps==MAX_AP_WARDRIVING) printbtm(20,0,"LIST FULL!");
		sprintf(renderbuf,"%i found",numaps);
		printbtm(10,0,renderbuf);

		// render::status line
		//             "01234567890123456789012345678901"
		printbtm( 0,23,"[UP/DN] Scroll   [B] Exit" ); 


		// render::scrollbar & stuff (we have 7 entries on screen, scrollbar height is 20 characters.)
		if(numaps<=7) {
			j=4; k=44;
		} else {
			j=(scrolltop*40)/numaps +4;
			k=((scrolltop+7)*40)/numaps +4;
		}
		printbtm( 30, 0,"\xb3\x18");
		printbtm( 30, 1,"\xc5\xc4");
		printbtm( 30,22,"\xc5\xc4");
		printbtm( 30,23,"\xb3\x19");

		for(i=2;i<22;i++) {
			if(i*2+1==j) { // top single piece
				printbtm( 30, i,"\xb3\xdc");
			} else if(i*2==k) { // bottom single piece
				printbtm( 30, i,"\xb3\xdf");
			} else 	if(i*2>=j && i*2<=k) {
				printbtm( 30, i,"\xb3\xdb");
			} else {
				printbtm( 30, i,"\xb3 ");
			}
		}
		
		// render::actual data
		//             "01234567890123456789012345678901"
		for(i=0;i<7;i++) {
			if(i+scrolltop>=numaps) break;
			sprintf(renderbuf,"ch%2i SSID:%20s",aplist[i+scrolltop].channel,aplist[i+scrolltop].ssid);
			printbtm( 0, 2+3*i,renderbuf ); 
			j='-';
			if(aplist[i+scrolltop].flags&WFLAG_APDATA_COMPATIBLE) j='+';
			if(aplist[i+scrolltop].flags&WFLAG_APDATA_EXTCOMPATIBLE) j='?';
			if(!(aplist[i+scrolltop].flags&WFLAG_APDATA_WPA)) { 
				//                "01234  56  78  9A  BC  DE F0123456789 ABCDEF "
				sprintf(renderbuf,"M%02X%02X%02X%02X%02X%02X  WEP%s C%c S%3i%%",aplist[i+scrolltop].mac[0],aplist[i+scrolltop].mac[1],
					aplist[i+scrolltop].mac[2],aplist[i+scrolltop].mac[3],aplist[i+scrolltop].mac[4],aplist[i+scrolltop].mac[5],
					aplist[i+scrolltop].flags&WFLAG_APDATA_WEP?"on ":"off",j,(aplist[i+scrolltop].rssi*100)/0xD0);
			} else {
				//                "01234  56  78  9A  BC  DE  F0123456789 ABCDEF "
				sprintf(renderbuf,"M%02X%02X%02X%02X%02X%02X  WPAon  C%c S%3i%%",aplist[i+scrolltop].mac[0],aplist[i+scrolltop].mac[1],
					aplist[i+scrolltop].mac[2],aplist[i+scrolltop].mac[3],aplist[i+scrolltop].mac[4],aplist[i+scrolltop].mac[5],
					j,(aplist[i+scrolltop].rssi*100)/0xD0);
			}

			printbtm( 0, 3+3*i,renderbuf ); 
			printbtm( 0, 4+3*i,"\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4" ); 

		}
		printbtm( 0, 1,"\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4" ); 
		printbtm( 0, 22,"\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4" ); 
		

		// check for new entries.
		for(i=0;i<Wifi_GetNumAP();i++) {
			if(WIFI_RETURN_OK!=Wifi_GetAPData(i,&ap)) continue;
			add=1;
			// search for AP in current list
			temp=aphashtabl[ap.macaddr[5]];
			while(temp) {
				k=1;
				for(j=0;j<6;j++) if(ap.macaddr[j]!=temp->mac[j]) k=0;
				if(k==1) { add=0; break; }
				temp=temp->next;
			}
			if(add) { // we do have to add the data.
				if(numaps==MAX_AP_WARDRIVING) continue;
				temp=aplist+numaps;
				//temp = &aplist[numaps];
				numaps++;
				temp->next=aphashtabl[ap.macaddr[5]];
				aphashtabl[ap.macaddr[5]]=temp;
				for(j=0;j<6;j++) temp->mac[j]=ap.macaddr[j];
			} // now, just update.
			temp->channel=ap.channel;
			temp->flags=ap.flags;
			temp->rssi=ap.rssi;
			for(j=0;j<ap.ssid_len;j++) temp->ssid[j]=ap.ssid[j];
			temp->ssid[j]=0;
		}
	}
	free(aplist);
	while(keysHeld()&KEY_TOUCH) WaitVbl();
}

#define MAX_KEYBOARD_STRING	1023
char keyboard_string[MAX_KEYBOARD_STRING+1];

#define KEYBOARD_NUMCHARS 47
const char * kb_lc_string = "`1234567890-=qwertyuiop[]\\asdfghjkl;'zxcvbnm,./";
const char * kb_uc_string = "~!@#$%^&*()_+QWERTYUIOP{}|ASDFGHJKL:\"ZXCVBNM<>?";
const char * kb_cl_string = "`1234567890-=QWERTYUIOP[]\\ASDFGHJKL;'ZXCVBNM,./";
const char * kb_template_lines[4] =  {
	"\xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3\x1b-\xb3",
	"\xb3-\x1a\xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3",
	"\xb3""CL \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3OK\xb3",
	"\xb3SHFT\xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3 \xb3SHF\xb3",
};
const int kb_char_y[KEYBOARD_NUMCHARS] = {
	0,0,0,0,0, 0,0,0,0,0, 0,0,0,1,1, 1,1,1,1,1, 1,1,1,1,1, 1,2,2,2,2, 2,2,2,2,2, 2,2,3,3,3, 3,3,3,3,3, 3,3
};
const int kb_char_x[KEYBOARD_NUMCHARS] = {
	1,3,5,7,9,11,13,15,17,19,21,23,25,
	4,6,8,10,12,14,16,18,20,22,24,26,28,
	5,7,9,11,13,15,17,19,21,23,25,
	6,8,10,12,14,16,18,20,22,24
};


char * EnterKeyboardText(const char * prompt, const char * originaltext, int (*UpdateFn)()) {
	int cursorpos,linelen,firstdrawchar,need_gen_keys,framecount;
	int i,charid;
	const char * srckb;
	char kbline[4][33];
	static int keystate =0;
	if(!prompt) prompt="Keyboard:";
	if(!originaltext) originaltext="";
	linelen=cursorpos=strlen(originaltext);
	firstdrawchar=0;
	strcpy(keyboard_string,originaltext);
	need_gen_keys=1;
	framecount=0;
	srckb=kb_lc_string;
	while(1) {
		WaitVbl();
		framecount++;
		ClsBtm();
		// process input
		if(keysDown()&KEY_LEFT) {
			if(cursorpos>0) cursorpos--;
		}
		if(keysDown()&KEY_RIGHT) {
			if(cursorpos<linelen) cursorpos++;
		}
		if(keysDown()&KEY_B) {
			if(cursorpos>0) {
				for(i=cursorpos;i<linelen;i++) {
					keyboard_string[i-1]=keyboard_string[i];
				}
				cursorpos--;
				linelen--;
				keyboard_string[linelen]=0;
			}
		}
		if(keysDown()&KEY_UP) {
			if(cursorpos<linelen) {
				if(keyboard_string[cursorpos]>=32 && keyboard_string[cursorpos]!=255) keyboard_string[cursorpos]++;
			}
		}
		if(keysDown()&KEY_DOWN) {
			if(cursorpos<linelen) {
				if(keyboard_string[cursorpos]>32) keyboard_string[cursorpos]--;
			}
		}
		if(keysDown()&KEY_START) {
			break; // accept string.
		}

		if(keysDown()&KEY_TOUCH) { // touch tapped - figure out what eet wants.
			charid=0; // no char yet.
			// process "special" regions.
			// "left" - center of (1,2) to center of (3,4)
			if(touchXY.px>=(1*8+4) && touchXY.px<(3*8+4) && touchXY.py>=(2*8+4) && touchXY.py<(4*8+4)) {
				if(cursorpos>0) cursorpos--;
			}
			// "right" - center of (28,2) to center of (30,4)
			if(touchXY.px>=(28*8+4) && touchXY.px<(30*8+4) && touchXY.py>=(2*8+4) && touchXY.py<(4*8+4)) {
				if(cursorpos<linelen) cursorpos++;
			}
			// touch in text area - move cursor
			if(touchXY.px>=(4*8) && touchXY.px<(28*8) && touchXY.py>=(2*8+4) && touchXY.py<(4*8+4)) {
				cursorpos=(touchXY.px/8)-4+firstdrawchar;
				if(cursorpos>linelen) cursorpos=linelen;
			}

			// "backspace" - center of (27,6) to center of (30,8)
			if(touchXY.px>=(27*8+4) && touchXY.px<(30*8+4) && touchXY.py>=(6*8+4) && touchXY.py<(8*8+4)) {
				if(cursorpos>0) {
					for(i=cursorpos;i<linelen;i++) {
						keyboard_string[i-1]=keyboard_string[i];
					}
					cursorpos--;
					linelen--;
					keyboard_string[linelen]=0;
				}
			}
			// "tab" - center of (1,8) to center of (4,10)
		//	if(touchXY.px>=(1*8+4) && touchXY.px<(4*8+4) && touchXY.py>=(8*8+4) && touchXY.py<(10*8+4)) {
		//
		//	}
			// "capslock" - center of (1,10) to center of (5,12)
			if(touchXY.px>=(1*8+4) && touchXY.px<(5*8+4) && touchXY.py>=(10*8+4) && touchXY.py<(12*8+4)) {
				keystate ^=2; need_gen_keys=1;
			}
			// "ok" - center of (27,10) to center of (30,12)
			if(touchXY.px>=(27*8+4) && touchXY.px<(30*8+4) && touchXY.py>=(10*8+4) && touchXY.py<(12*8+4)) {
				break;
			}
			// "lshft" - center of (1,12) to center of (6,14)
			if(touchXY.px>=(1*8+4) && touchXY.px<(6*8+4) && touchXY.py>=(12*8+4) && touchXY.py<(14*8+4)) {
				keystate ^=1; need_gen_keys=1;
			}
			// "rshft" - center of (26,12) to center of (30,14)
			if(touchXY.px>=(26*8+4) && touchXY.px<(30*8+4) && touchXY.py>=(12*8+4) && touchXY.py<(14*8+4)) {
				keystate ^=1; need_gen_keys=1;
			}
			// "ctrl" - center of (28,12) to center of (30,14)
	//		if(touchXY.px>=(28*8+4) && touchXY.px<(30*8+4) && touchXY.py>=(12*8+4) && touchXY.py<(14*8+4)) {
	//
	//		}
			// "space" - center of (9,14) to center of (21,16)
			if(touchXY.px>=(9*8+4) && touchXY.px<(21*8+4) && touchXY.py>=(14*8+4) && touchXY.py<(16*8+4)) {
				charid=32;
			}
			if(!UpdateFn) {
				// "cancel" - center of (1,20) to center of (16,22)
				if(touchXY.px>=(1*8+4) && touchXY.px<(16*8+4) && touchXY.py>=(20*8+4) && touchXY.py<(22*8+4)) {
					while(keysHeld()&KEY_TOUCH) WaitVbl();
					return NULL;
				}
				// "accept" - center of (17,20) to center of (30,22)
				if(touchXY.px>=(17*8+4) && touchXY.px<(30*8+4) && touchXY.py>=(20*8+4) && touchXY.py<(22*8+4)) {
					break;
				}
			}

			// process "key" regions (keyboard starts at (8,48) technically)
			for(i=0;i<KEYBOARD_NUMCHARS;i++) {
				if(touchXY.px>=(12-8+kb_char_x[i]*8) && touchXY.px<(12+8+kb_char_x[i]*8) && touchXY.py>=(60-8+kb_char_y[i]*16) && touchXY.py<(60+8+kb_char_y[i]*16)) {
					charid=srckb[i]; break;
				}
			}
			if(charid!=0) {
				if(linelen<MAX_KEYBOARD_STRING) {
					for(i=linelen;i>cursorpos;i--) {
						keyboard_string[i]=keyboard_string[i-1];
					}
					keyboard_string[cursorpos]=charid;
					cursorpos++;linelen++;
					keyboard_string[linelen]=0;
				}
				if(keystate==1) { keystate=0; need_gen_keys=1; }
			}
		}


		// draw::prompt & text
		//             "01234567890123456789012345678901"
		btm_drawrect2(0,0,31,23);
		printbtm(16-(strlen(prompt)>>1),0,prompt);

		if((cursorpos-1)<firstdrawchar) firstdrawchar=cursorpos-1;
		if(cursorpos>(firstdrawchar+22)) firstdrawchar=cursorpos-22;
		if(firstdrawchar<0) firstdrawchar=0;
		//             "01234567890123456789012345678901" // space of 24 characters to draw string
		printbtm( 1, 2,"\xda\xc4\xc2\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc2\xc4\xbf");
		printbtm( 1, 3,"\xb3 \xb3                        \xb3 \xb3");
		printbtm( 1, 4,"\xc0\xc4\xc1\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc1\xc4\xd9");
		printbtmn( 4, 3,keyboard_string+firstdrawchar,24);
		if(framecount&32) printbtm(4+cursorpos-firstdrawchar,3,"_");
		if(cursorpos>0) printbtm(2,3,"\x1b");
		if(cursorpos<linelen) printbtm(29,3,"\x1a");

		// draw::keyboard
		if(need_gen_keys) {
			need_gen_keys=0;
			switch(keystate) {
			case 0: // lowercase
				srckb=kb_lc_string; break;
			case 1: // shift
			case 3: // caps+shift
				srckb=kb_uc_string; break;
			case 2: // caps
				srckb=kb_cl_string; break;
			}
			for(i=0;i<4;i++) {
				strcpy(kbline[i],kb_template_lines[i]);
			}
			for(i=0;i<KEYBOARD_NUMCHARS;i++) {
				kbline[kb_char_y[i]][kb_char_x[i]]=srckb[i];
			}
		}

		printbtm( 1, 6,"\xda\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc2\xc4\xc4\xbf");
		printbtm( 1, 7,kbline[0]);
		printbtm( 1, 8,"\xc3\xc4\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc4\xb4");
		printbtm( 1, 9,kbline[1]);
		printbtm( 1,10,"\xc3\xc4\xc4\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc4\xb4");
		printbtm( 1,11,kbline[2]);
		printbtm( 1,12,"\xc3\xc4\xc4\xc4\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc2\xc1\xc4\xc4\xb4");
		printbtm( 1,13,kbline[3]);
		printbtm( 1,14,"\xc0\xc4\xc4\xc2\xc4\xc1\xc4\xc1\xc2\xc1\xc4\xc1\xc4\xc1\xc4\xc1\xc4\xc1\xc4\xc1\xc2\xc1\xc4\xc1\xc4\xc1\xc4\xc4\xc4\xd9");
		printbtm( 1,15,"   \xb3""CTRL\xb3           \xb3         ");
		printbtm( 1,16,"   \xc0\xc4\xc4\xc4\xc4\xc1\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xc4\xd9         ");

		// FOR REFERENCE: (original keyboard, hehe ;)
		//             "01234567890123456789012345678901"
		/*printbtm( 0, 6," ------------------------------");
		printbtm( 0, 7," |`|1|2|3|4|5|6|7|8|9|0|-|=|<-|");
		printbtm( 0, 8," ------------------------------");
		printbtm( 0, 9," |->|q|w|e|r|t|y|u|i|o|p|[|]|\|");
		printbtm( 0,10," ------------------------------");
		printbtm( 0,11," |CL |a|s|d|f|g|h|j|k|l|;|'|OK|");
		printbtm( 0,12," ------------------------------");
		printbtm( 0,13," |SHFT|z|x|c|v|b|n|m|,|.|/|SHF|");
		printbtm( 0,14," ------------------------------");
		printbtm( 0,15,"         |           |         ");
		printbtm( 0,16,"         -------------         ");*/

		if(UpdateFn) {
			if((i=UpdateFn())) return NULL;
		} else {
			btm_drawbutton(1,20,16,22,0,"Cancel");
			btm_drawbutton(17,20,30,22,0,"Accept");
		}

	}
	while(keysHeld()&KEY_TOUCH) WaitVbl();

	return keyboard_string;
}
unsigned long EnterIPAddress(const char * prompt, unsigned long originalip, int (*UpdateFn)()) {
	char ipaddr_buf[32];
	char * dest;
	int i,j;
	sprintf(ipaddr_buf,"%i.%i.%i.%i",(int)((originalip)&255),(int)((originalip>>8)&255),(int)((originalip>>16)&255),(int)((originalip>>24)&255));
	dest=EnterKeyboardText(prompt,ipaddr_buf,UpdateFn);
	if(!dest) return originalip;
	i=0;j=0;
	originalip=0;
	while(j<4) {
		if(*dest=='.' || *dest==0) {
			originalip = (originalip >> 8) | (i<<24);
			i=0;
			j++;
			if(*dest==0 && j<4) {
				originalip = originalip >> ((4-j)*8);
				j=4;
			}
		} else {
			if(*dest>='0' && *dest<='9') {
				i=i*10+(*dest-'0');
			}
		}
		dest++;
	}
	return originalip;
}

int EnterWEPKey(const char * prompt, unsigned char * wepkey, int byte_length, int (*UpdateFn)()) {
	char wepkey_buf[64];
	char * dest;
	int i,j;
	for(i=0;i<byte_length;i++) {
		j=wepkey[i]>>4;
		if(j<10) wepkey_buf[i*2]=j+'0'; else wepkey_buf[i*2]=j+'A'-10;
		j=wepkey[i]&15;
		if(j<10) wepkey_buf[i*2+1]=j+'0'; else wepkey_buf[i*2+1]=j+'A'-10;
	}
	wepkey_buf[i*2]=0;
	dest = EnterKeyboardText(prompt,wepkey_buf,UpdateFn);
	if(!dest) return -1;
	i=0;
	while(*dest) {
		j=0;
		if(*dest>='0' && *dest<='9') j=*dest-'0';
		if(*dest>='A' && *dest<='F') j=*dest-'A'+10;
		if(*dest>='a' && *dest<='f') j=*dest-'a'+10;

		if(i&1) wepkey[i/2] = (wepkey[i/2]&0xF0)|j; else wepkey[i/2] = (wepkey[i/2]&0xF)|(j<<4);
		i++;
		dest++;
	}
	return i;
}

void Do_Play_KeyboardTest()	{
	char buffer[255];
	char * ret;
	ret=EnterKeyboardText("Keyboard Test!", "Hi!", 0);
	if(ret) {
		sprintf(buffer,"You entered: '%s'",ret);
	} else {
		strcpy(buffer,"You Canceled the keyboard test!");
	}
	printtop(buffer);
}

#define SCROLL_DELAY 32
int Do_SelectListAP() {
	char printbuf[32];
	int i,j,selected,scrolltop,scrolldel,numap;
	int retvalue;
	Wifi_AccessPoint ap;
	Wifi_ScanMode();
	selected=0;
	scrolltop=0;
	scrolldel=100;
	retvalue=1;
	while(1) {
		WaitVbl();
		ClsBtm();
		numap=Wifi_GetNumAP();
		// process input
		if(keysHeld()&KEY_TOUCH) { // update selection
			if(touchXY.px>16 && touchXY.px<240 && touchXY.py>8 && touchXY.py<160) {
				if(touchXY.py<16) { // scroll up
					if(scrolldel==0) {
						scrolltop--;
						if(scrolltop<0) scrolltop=0;
						scrolldel=SCROLL_DELAY;
						if(selected>=(scrolltop+17)) selected--;
					}
					scrolldel--;
				} else if(touchXY.py>152) { // scroll down
					if(scrolldel==0) {
						scrolltop++;
						if(scrolltop>numap-17) scrolltop=numap-17;
						scrolldel=SCROLL_DELAY;
					}
					scrolldel--;
				} else {
					j=(touchXY.py-16)/8;
					if(scrolltop+j<numap) selected=scrolltop+j;
					if(selected<=scrolltop) selected=scrolltop;
				}
			}
		} else {
			scrolldel=0;
		}
		if(keysDown()&KEY_TOUCH) { // check touch "buttons"
			// "cancel" - center of (1,20) to center of (16,22)
			if(touchXY.px>=(1*8+4) && touchXY.px<(16*8+4) && touchXY.py>=(20*8+4) && touchXY.py<(22*8+4)) {
				retvalue=0;
				break;
			}
			// "accept" - center of (17,20) to center of (30,22)
			if(touchXY.px>=(17*8+4) && touchXY.px<(30*8+4) && touchXY.py>=(20*8+4) && touchXY.py<(22*8+4)) {
				Wifi_GetAPData(selected,&global_connectAP);
				break;
			}

		}
	
		if(keysDown()&KEY_UP) {
			if(selected>0) selected--;
			if((selected-1) < scrolltop && selected!=0) scrolltop=selected-1;
		}
		if(keysDown()&KEY_DOWN) {
			if(selected<(numap-1)) selected++;
			if((selected-16)> scrolltop) scrolltop = selected-16;
		}
		if(keysDown()&(KEY_A|KEY_START)) {
			Wifi_GetAPData(selected,&global_connectAP);
			break;
		}
		if(keysDown()&(KEY_B)) {
			retvalue=0;
			break;
		}

		if(scrolltop>numap-17) scrolltop=numap-17;
		if(scrolltop<0) scrolltop=0;

		// update display
		printbtm(0,0,"AP List");
		btm_drawrect1(1,1,30,19);
		btm_drawbutton(1,20,16,22,0,"Cancel");
		btm_drawbutton(17,20,30,22,0,"Accept");
		
		printbtm(0,2+selected-scrolltop,">");
		printbtm(31,2+selected-scrolltop,"<");
		for(i=0;i<17;i++) {
			if(i+scrolltop>=numap) break;
			if(Wifi_GetAPData(i+scrolltop,&ap)==WIFI_RETURN_OK) {
				j='-';
				if(ap.flags&WFLAG_APDATA_COMPATIBLE) j='+';
				if(ap.flags&WFLAG_APDATA_EXTCOMPATIBLE) j='?';
				// we have 28 characters to work with
				//               "01 23456 789ABCDEF0123456789AB"
				sprintf(printbuf,"C%cS%3i%%:%20s",j,(ap.rssi*100)/0xD0,ap.ssid);
				printbtm(2,2+i,printbuf);
			}
		}
	}
	Wifi_DisableWifi();
	while(keysHeld()&KEY_TOUCH) WaitVbl();
	WaitVbl();WaitVbl();
	return retvalue;
}
const char * wepmodes[] = {
	"off","64bit","?","128bit"
};
const int wepkeylen[] = {
	0,5,0,13
};
const char * dhcpmodes[] = {
	"Static IP","DHCP IP, Manual DNS", "Static IP, DHCP DNS", "DHCP IP and DNS"
};
int Do_ManualConfigureAP() {
	char tmp[36];
	char * temp;
	int i;
	while(1) {
		WaitVbl();
		ClsBtm();
		// handle input
		if(keysDown()&KEY_TOUCH) {
			// "edit ssid" - center of (21,1) to center of (26,3)
			if(touchXY.px>=(21*8+4) && touchXY.px<(26*8+4) && touchXY.py>=(1*8+4) && touchXY.py<(3*8+4)) {
				temp=EnterKeyboardText("Change SSID",global_connectAP.ssid,0);
				if(temp){
					temp[32]=0;
					i=strlen(temp);
					if(strcmp(global_connectAP.ssid,temp)) {
						strcpy(global_connectAP.ssid,temp);
						global_connectAP.ssid_len=i;
						for(i=0;i<6;i++) global_connectAP.macaddr[i]=0; // clear mac if changing ssid.
						global_connectAP.base_rates[0]=0; // clear rates if changing ssid.
					}
				}
			}
			// "ch%i" - center of (26,1) to center of (31,3)
			if(touchXY.px>=(26*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(1*8+4) && touchXY.py<(3*8+4)) {
				global_connectAP.channel++;
				if(global_connectAP.channel>13) global_connectAP.channel=1;
			}
			// "wepmode" - center of (0,3) to center of (11,5)
			if(touchXY.px>=(0*8+4) && touchXY.px<(11*8+4) && touchXY.py>=(3*8+4) && touchXY.py<(5*8+4)) {
				global_wepmode++;
				if(global_wepmode==2) global_wepmode++;
				if(global_wepmode>3) global_wepmode=0;
			}
			// "wepkeyid" - center of (11,3) to center of (21,5)
			if(touchXY.px>=(11*8+4) && touchXY.px<(21*8+4) && touchXY.py>=(3*8+4) && touchXY.py<(5*8+4)) {
				global_wepkeyid++;
				if(global_wepkeyid>3) global_wepkeyid=0;
			}
			// "wepeditkey" - center of (21,3) to center of (31,5)
			if(touchXY.px>=(21*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(3*8+4) && touchXY.py<(5*8+4)) {
				switch(EnterWEPKey("Change WEP key",global_wepkeys[0],wepkeylen[global_wepmode],0)) {
					case 0: global_wepmode=0; break;
					case 10: global_wepmode=1; break;
					case 26: global_wepmode=3; break;
				}
			}
			// "iptype" - center of (8,5) to center of (31,7)
			if(touchXY.px>=(8*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(5*8+4) && touchXY.py<(7*8+4)) {
				global_dhcp++;
				if(global_dhcp>3) global_dhcp=0;
			}
			// "editipaddr" - center of (26,7) to center of (31,9)
			if(touchXY.px>=(26*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(7*8+4) && touchXY.py<(9*8+4)) {
				global_ipaddr=EnterIPAddress("Change IP Address",global_ipaddr,0);
			}
			// "editgateway" - center of (26,9) to center of (31,11)
			if(touchXY.px>=(26*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(9*8+4) && touchXY.py<(11*8+4)) {
				global_gateway=EnterIPAddress("Change Gateway Address",global_gateway,0);
			}
			// "editsnmask" - center of (26,11) to center of (31,13)
			if(touchXY.px>=(26*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(11*8+4) && touchXY.py<(13*8+4)) {
				global_snmask=EnterIPAddress("Change Subnet Mask",global_snmask,0);
			}
			// "editdns1" - center of (26,13) to center of (31,15)
			if(touchXY.px>=(26*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(13*8+4) && touchXY.py<(15*8+4)) {
				global_dns1=EnterIPAddress("Change DNS 1 Address",global_dns1,0);
			}
			// "editdns2" - center of (26,15) to center of (31,17)
			if(touchXY.px>=(26*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(15*8+4) && touchXY.py<(17*8+4)) {
				global_dns2=EnterIPAddress("Change DNS 2 Address",global_dns2,0);
			}
			// "cancel" - center of (0,21) to center of (7,23)
			if(touchXY.px>=(0*8+4) && touchXY.px<(7*8+4) && touchXY.py>=(21*8+4) && touchXY.py<(23*8+4)) {
				return 0;
			}
			// "findnewap" - center of (7,21) to center of (23,23)
			if(touchXY.px>=(7*8+4) && touchXY.px<(23*8+4) && touchXY.py>=(21*8+4) && touchXY.py<(23*8+4)) {
				Do_SelectListAP();
			}
			// "connect" - center of (23,21) to center of (31,23)
			if(touchXY.px>=(23*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(21*8+4) && touchXY.py<(23*8+4)) {
				return 1;
			}

		}

		// update display
		printbtm(0,0,"Edit/Check Settings");
		btm_pddrawbutton(21,1,26,3,0,"Edit");
		sprintf(tmp,"ch%i",global_connectAP.channel);
		btm_pddrawbutton(26,1,31,3,0,tmp);
		printbtm(0,2,"SSID:");
		printbtmn(5,2,global_connectAP.ssid,16);
	
		sprintf(tmp,"WEP:%s",wepmodes[global_wepmode]);
		btm_pddrawbutton(0,3,11,5,0,tmp);
		sprintf(tmp,"KeyID:%i",global_wepkeyid);
		btm_pddrawbutton(11,3,21,5,0,tmp);
		btm_pddrawbutton(21,3,31,5,0,"Edit Key");

		printbtm(0,6,"TCP/IP:");
		btm_pddrawbutton(8,5,31,7,0,dhcpmodes[global_dhcp]);
		// lines 10 - 20 allocated to static IP/DNS.  Button merging will be required.
		
		btm_pddrawbutton(26,7,31,9,0,"Edit");
		btm_pddrawbutton(26,9,31,11,0,"Edit");
		btm_pddrawbutton(26,11,31,13,0,"Edit");
		btm_pddrawbutton(26,13,31,15,0,"Edit");
		btm_pddrawbutton(26,15,31,17,0,"Edit");
		// max length of IP address is 15 characters. As such, the initial text must be <=11 characters to fit anything possible.
		printbtm(0,8,"IP Address:"); // 11
		sprintf(tmp,"%i.%i.%i.%i",(int)(global_ipaddr&255),(int)((global_ipaddr>>8)&255),(int)((global_ipaddr>>16)&255),(int)((global_ipaddr>>24)&255));
		printbtm(11,8,tmp);
		printbtm(0,10,"Gateway:"); // 8
		sprintf(tmp,"%i.%i.%i.%i",(int)(global_gateway&255),(int)((global_gateway>>8)&255),(int)((global_gateway>>16)&255),(int)((global_gateway>>24)&255));
		printbtm(8,10,tmp);
		printbtm(0,12,"SubnetMask:"); // 11
		sprintf(tmp,"%i.%i.%i.%i",(int)(global_snmask&255),(int)((global_snmask>>8)&255),(int)((global_snmask>>16)&255),(int)((global_snmask>>24)&255));
		printbtm(11,12,tmp);
		printbtm(0,14,"DNS 1:"); // 6
		sprintf(tmp,"%i.%i.%i.%i",(int)(global_dns1&255),(int)((global_dns1>>8)&255),(int)((global_dns1>>16)&255),(int)((global_dns1>>24)&255));
		printbtm(6,14,tmp);
		printbtm(0,16,"DNS 2:"); // 6
		sprintf(tmp,"%i.%i.%i.%i",(int)(global_dns2&255),(int)((global_dns2>>8)&255),(int)((global_dns2>>16)&255),(int)((global_dns2>>24)&255));
		printbtm(6,16,tmp);
		
		btm_pddrawbutton(0,21,7,23,0,"Cancel");
		btm_pddrawbutton(23,21,31,23,0,"Connect");
		btm_pddrawbutton(7,21,23,23,0,"Find New AP");

	}
}

const char * assocstatus_strings[] = {
	"ASSOCSTATUS_DISCONNECTED", // not *trying* to connect
	"ASSOCSTATUS_AUTHENTICATING", // connecting...
	"ASSOCSTATUS_ASSOCIATING", // connecting...
	"ASSOCSTATUS_ASSOCIATED",	// Connected!
	"ASSOCSTATUS_CANNOTCONNECT", // error in connecting...
};
int Do_ConnectAP() {
	int i,j,state,delay;
	int anim[3],dir[3],rate[3];
	//char tmpbuf[64];
	delay=0;
	Wifi_DisconnectAP();
	// set DHCP/Static IP settings...
	Wifi_SetStaticIP(global_ipaddr,global_gateway,global_snmask,global_dns1,global_dns2);
	// set startup state...
	anim[0]=1;
	anim[1]=7;
	anim[2]=15;
	dir[0]=dir[2]=1;
	dir[1]=-1;
	rate[0]=rate[1]=rate[2]=0;
	j=0;
	for(i=0;i<6;i++) j|=global_connectAP.macaddr[i];
	if(j==0) state=0; else state=5;
	while(1) {
		WaitVbl();
		ClsBtm();
		// Process Input
		if(keysDown()&KEY_B) {
			Wifi_DisconnectAP();
			Wifi_DisableWifi();
			return 0;
		}
		if(keysDown()&KEY_TOUCH) {
			// "cancel" - center of (16,21) to center of (31,23)
			if(touchXY.px>=(16*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(21*8+4) && touchXY.py<(23*8+4)) {
				Wifi_DisconnectAP();
				Wifi_DisableWifi();
				return 0;
			}
		}

		// Draw info
		printbtm(0,0,"Connecting to AP...");
		btm_drawbutton(16,21,31,23,0,"Cancel");
		printbtm(0,23,"[B]Cancel");

		// draw animation doodad.
		for(i=0;i<3;i++) {
			rate[i]+=256;
			if(rate[i]>(i*350+600)) {
				rate[i]-=(i*350+600);
				anim[i]+=dir[i];
				if(anim[i]==22) dir[i]=-1;
				if(anim[i]==0) dir[i]=1;
			}
			btm_pddrawrect2(anim[i],15+i,anim[i]+9,18+i); 
		}


		// draw state info & processing.
		switch(state) {
		case 0: // lookup mac address
			//           "0123456789ABCDEF0123456789ABCDEF"
			printbtm(0,2,"Attempting to find AP's MAC...");
			printbtm(0,3,"(Not implemented yet, please go");
			printbtm(0,4," back and pick an AP with the");
			printbtm(0,5," AP list, don't change the SSID)");
			break;

		case 5: // connect to AP
			state=6;
			Wifi_ConnectAP(&global_connectAP,global_wepmode,global_wepkeyid,global_wepkeys[0]);
		case 6:
			//           "0123456789ABCDEF0123456789ABCDEF"
			printbtm(0,2,"Connecting to Access Point...");
			printbtm(0,4,assocstatus_strings[Wifi_AssocStatus()]);
			if(Wifi_AssocStatus() == ASSOCSTATUS_ASSOCIATED) state=30;
			break;		

		case 30:
			delay=60;
			state=31;
		case 31:
			printbtm(0,2,"Connected!");
			if(!(delay--)) return 1;
			break;

		}


	}

}

int Do_ConfigureWifi() {
	int i,state,touchstate;
	printtop("Do_ConfigureWifi");
	Wifi_EnableWifi();
	Wifi_ScanMode();
	state=0;
	while(1) {
		state=0;touchstate=-1;
		while(1) {
			WaitVbl();
			ClsBtm();
			if(keysUp()&KEY_TOUCH && touchstate==state) { // just released
				break; // state confirmed.
			}
			touchstate=-1;
			if(keysHeld()&KEY_TOUCH) { // update touchstate
				if(touchXY.px>16 && touchXY.px<240) {
					for(i=0;i<NUM_CONF_MENU_ITEMS;i++) {
						if(touchXY.py>16+32*i && touchXY.py<40+32*i) { touchstate=i; }
					}
				}
				if(touchstate!=-1) state=touchstate;
			}
			if(keysDown()&KEY_UP) {
				if(state==0) state=NUM_CONF_MENU_ITEMS;
				state--;
			}
			if(keysDown()&KEY_DOWN) {
				state++;
				if(state==NUM_CONF_MENU_ITEMS) state=0;
			}
			if(keysDown()&(KEY_A | KEY_START)) {
				break; // state confirmed
			}
			if(keysDown()&(KEY_B)) {
				state=NUM_CONF_MENU_ITEMS-1;
				break; // abort! abort!
			}

			for(i=0;i<NUM_CONF_MENU_ITEMS;i++) {
				btm_drawbutton(3,1+i*4,28,3+i*4,i==state,conf_mnuitems[i]);
			}
			printbtm(0,23,"[UP/DN]Move[A/Touch]Sel[B]Return");
		}
		while(keysHeld()&KEY_TOUCH) WaitVbl();
		if(state==(NUM_CONF_MENU_ITEMS-1)) return 1;
		switch(state) {
			case 0: // search
				if(!Do_SelectListAP()) {
					printtop("Canceled Select AP");
					break;
				}
			case 1: // manual
				if(Do_ManualConfigureAP()) {
					// connect to ap
					if(Do_ConnectAP()) {
						return 0; // zero is no-error in this case :)
					}
				}
				break;
			case 2: // WFC
			break;
		}		
	}
}
char rcvdbuf[4096];
char sendbuf[4096];
void Do_Play_UDPTest() {
	int sock;
	int portnum;
	int rcvd,i,j,lastmsglen;
	unsigned long destip,srcip;
	struct sockaddr_in sain, sain_;
	char * tmp;
	char buf[64];
	strcpy(sendbuf,"Test String!");
	tmp = EnterKeyboardText("Pick UDP port to use","8888",0);
	if(!tmp) return;
	portnum=atoi(tmp);

	sock=socket(AF_INET,SOCK_DGRAM,0);
	sain.sin_family=AF_INET;
	sain.sin_port=htons(portnum);
	sain.sin_addr.s_addr=INADDR_ANY;
	bind(sock,(struct sockaddr *) &sain,sizeof(sain));
	rcvd=0;
	i=1;
	ioctl(sock,FIONBIO,&i);

	destip=Wifi_GetIP();
	lastmsglen=0;

	while(1) {
		WaitVbl();
		ClsBtm();
		// Process input
		if(keysDown()&KEY_B) {
			break;
		}
		if(keysDown()&KEY_TOUCH) {
			// "edit" - center of (24,16) to center of (31,18)
			if(touchXY.px>=(24*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(16*8+4) && touchXY.py<(18*8+4)) {
				destip=EnterIPAddress("Change Dest IP",destip,0);
			}
			// "edit message" - center of (0,18) to center of (24,20)
			if(touchXY.px>=(0*8+4) && touchXY.px<(24*8+4) && touchXY.py>=(18*8+4) && touchXY.py<(20*8+4)) {
				tmp=EnterKeyboardText("Change Message Text",sendbuf,0);
				if(tmp) strcpy(sendbuf,tmp);
			}
			// "send" - center of (24,18) to center of (31,20)
			if(touchXY.px>=(24*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(18*8+4) && touchXY.py<(20*8+4)) {
				sain.sin_family=AF_INET;
				sain.sin_port=htons(portnum);
				sain.sin_addr.s_addr=destip;
				sendto(sock,sendbuf,strlen(sendbuf),0,(struct sockaddr *)&sain,sizeof(sain));
			}
			// "exit" - center of (24,21) to center of (31,23)
			if(touchXY.px>=(24*8+4) && touchXY.px<(31*8+4) && touchXY.py>=(21*8+4) && touchXY.py<(23*8+4)) {
				break;
			}
		}


		// process incoming data...
		i=sizeof(sain_);
		if((j=recvfrom(sock,rcvdbuf,4095,0,(struct sockaddr *)&sain_,&i))!=-1) { lastmsglen=j; rcvd++; }
		
		// display info
		printbtm(0,0,"UDP Experimenter");
		btm_pddrawrect2(0,1,31,15);
		printbtm(2,1,"Last Received");
		sprintf(buf,"%i Received",rcvd);
		printbtm(29-strlen(buf),15,buf);
		if(rcvd) {
			srcip=sain_.sin_addr.s_addr;
			sprintf(buf,"From:%i.%i.%i.%i:%i",(srcip)&255,(srcip>>8)&255,(srcip>>16)&255,(srcip>>24)&255,ntohs(sain_.sin_port));
			printbtm(1,2,buf);
			sprintf(buf,"Msg Len: %i",lastmsglen);
			printbtm(1,3,buf);
			rcvdbuf[lastmsglen]=0;
			// have rows 4-14 to display msg- 11 lines
			for(i=0;i<11;i++) {
				if(i*30<lastmsglen) printbtmn(1,4+i,rcvdbuf+i*30,30);
			}
		}

		sprintf(buf,"DestIP:%i.%i.%i.%i",(int)(destip&255),(int)((destip>>8)&255),(int)((destip>>16)&255),(int)((destip>>24)&255));
		printbtm(1,17,buf);
		sprintf(buf,"Msglen: %ib",strlen(sendbuf));
		printbtm(29-strlen(buf),20,buf);
		btm_pddrawbutton(24,16,31,18,0,"Edit");
		btm_pddrawbutton(0,18,24,20,0,"Edit Message");
		btm_pddrawbutton(24,18,31,20,0,"Send");

		btm_pddrawbutton(24,21,31,23,0,"Exit");
		btm_pddrawrect2(0,16,31,20);
		printbtm(2,16,"Test Send");
		sprintf(buf,"Msglen: %ib",strlen(sendbuf));
		printbtm(23-strlen(buf),20,buf);

		printbtm(0,23,"[B]Exit");

	}
	close(sock);
}

